import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from SALib.analyze.morris import analyze as morris_analyze
from SALib.analyze.sobol import analyze as sobol_analyze
from SALib.sample.morris import sample as morris_sample
from SALib.sample.sobol import sample as sobol_sample
from SALib.util import ResultDict

from autoemulate.core.plotting import display_figure
from autoemulate.core.types import DistributionLike, NumpyLike, TensorLike
from autoemulate.data.utils import ConversionMixin
from autoemulate.emulators.base import Emulator

logger = logging.getLogger("autoemulate")


class SensitivityAnalysis(ConversionMixin):
    """Global sensitivity analysis."""

    def __init__(
        self,
        emulator: Emulator,
        x: TensorLike | None = None,
        problem: dict | None = None,
    ):
        """
        Initialize the SensitivityAnalysis.

        Parameters
        ----------
        emulator: Emulator
            Fitted emulator.
        x: InputLike | None
            Simulator input parameter values.
        problem: dict | None
            The problem definition dictionary. If None, the problem is generated
            from x using minimum and maximum values of the features as bounds.
            The dictionary should contain:
                - 'num_vars': Number of input variables (int)
                - 'names': List of variable names (list of str)
                - 'bounds': List of [min, max] bounds for each variable (list of lists)
                - 'output_names': Optional list of output names (list of str)

            Example::
                problem = {
                    "num_vars": 2,
                    "names": ["x1", "x2"],
                    "bounds": [[0, 1], [0, 1]],
                    "output_names": ["y1", "y2"],  # optional
                }
        """
        if problem is not None:
            problem = self._check_problem(problem)
        elif x is not None:
            problem = self._generate_problem(x)
        else:
            msg = "Either problem or x must be provided."
            raise ValueError(msg)

        self.emulator = emulator
        self.problem = problem

    @staticmethod
    def _check_problem(problem: dict) -> dict:
        """Check that the problem definition is valid."""
        if not isinstance(problem, dict):
            msg = "problem must be a dictionary."
            raise ValueError(msg)

        if "num_vars" not in problem:
            msg = "problem must contain 'num_vars'."
            raise ValueError(msg)
        if "names" not in problem:
            msg = "problem must contain 'names'."
            raise ValueError(msg)
        if "bounds" not in problem:
            msg = "problem must contain 'bounds'."
            raise ValueError(msg)

        if len(problem["names"]) != problem["num_vars"]:
            msg = "Length of 'names' must match 'num_vars'."
            raise ValueError(msg)
        if len(problem["bounds"]) != problem["num_vars"]:
            msg = "Length of 'bounds' must match 'num_vars'."
            raise ValueError(msg)

        return problem

    @staticmethod
    def _generate_problem(x: TensorLike) -> dict:
        """
        Generate a problem definition from a design matrix.

        Parameters
        ----------
        x: TensorLike
            Simulator input parameter values [n_samples, n_parameters].

        Returns
        -------
        dict
            Dictionary with keys ["num_vars", "names", "bounds"]
        """
        if x.ndim == 1:
            msg = "x must be a 2D array."
            raise ValueError(msg)

        return {
            "num_vars": x.shape[1],
            "names": [f"x{i + 1}" for i in range(x.shape[1])],
            "bounds": [
                [x[:, i].min().item(), x[:, i].max().item()] for i in range(x.shape[1])
            ],
        }

    def _sample(self, method: str, N: int) -> NumpyLike:
        if method == "sobol":
            # Saltelli sampling
            return sobol_sample(self.problem, N)
        if method == "morris":
            # vanilla Morris (1991) sampling
            return morris_sample(self.problem, N)
        msg = f"Unknown method: {method}. Must be 'sobol' or 'morris'."
        raise ValueError(msg)

    def _predict(self, param_samples: NumpyLike) -> NumpyLike:
        """
        Make predictions with emulator for parameter samples.

        Parameters
        ----------
        param_samples: NumpyLike
            Array of parameter samples.

        Returns
        -------
        NumpyLike
            Array of emulator predictions.
        """
        param_tensor = self._convert_to_tensors(param_samples)
        assert isinstance(param_tensor, TensorLike)
        y_pred = self.emulator.predict(param_tensor)

        # handle types, convert to numpy
        if isinstance(y_pred, TensorLike):
            y_pred_np, _ = self._convert_to_numpy(y_pred)
        elif isinstance(y_pred, DistributionLike):
            y_pred_np, _ = self._convert_to_numpy(y_pred.mean.float())
        else:
            msg = "Emulator has to return Tensor or Distribution"
            raise ValueError(msg)

        return y_pred_np

    def _get_output_names(self, num_outputs: int) -> list[str]:
        """Get output names from the problem definition or generate default names."""
        # check if output_names is given
        if "output_names" not in self.problem:
            output_names = [f"y{i + 1}" for i in range(num_outputs)]
        elif isinstance(self.problem["output_names"], list):
            output_names = self.problem["output_names"]
        else:
            msg = "'output_names' must be a list of strings."
            raise ValueError(msg)

        return output_names

    def run(
        self,
        method: str = "sobol",
        n_samples: int = 1024,
        conf_level: float = 0.95,
    ) -> pd.DataFrame:
        """
        Perform global sensitivity analysis on a fitted emulator.

        Parameters
        ----------
        method: str
            The sensitivity analysis method to perform, one of ["sobol", "morris"].
        n_samples: int
            Number of samples to generate for the analysis. Higher values give more
            accurate results but increase computation time. Default is 1024.
        conf_level: float
            Confidence level (between 0 and 1) for calculating confidence intervals
            of the Sobol sensitivity indices. Default is 0.95 (95% confidence).

        Returns
        -------
        pandas.DataFrame
            DataFrame with columns:
                - 'parameter': Input parameter name
                - 'output': Output variable name
            If using Sobol, the columns include:
                - 'index': S1, S2 or ST (first, second, and total order sensitivity)
                - 'confidence': confidence intervals for each index
            If using Morris, the columns include:
                - 'mu': mean of the distribution of elementary effects
                - 'mu_star': mean of the distribution of absolute value
                - 'sigma': standard deviation of the distribution, used as indication of
                    interactions between parameters
                - 'mu_star_conf: boostrapped confidence interval

        Notes
        -----
        The Sobol method requires N * (2D + 2) model evaluations, where D is the number
        of input parameters. For example, with N=1024 and 5 parameters, this requires
        12,288 evaluations. The Morris method requires far fewer computations.
        """
        logger.debug(
            "Running sensitivity analysis with method=%s, n_samples=%d, conf_level=%s",
            method,
            n_samples,
            conf_level,
        )

        if method not in ["sobol", "morris"]:
            msg = f"Unknown method: {method}. Must be 'sobol' or 'morris'."
            raise ValueError(msg)

        param_samples = self._sample(method, n_samples)
        y = self._predict(param_samples)
        output_names = self._get_output_names(y.shape[1])

        results = {}
        for i, name in enumerate(output_names):
            if method == "sobol":
                Si = sobol_analyze(self.problem, y[:, i], conf_level=conf_level)
            elif method == "morris":
                Si = morris_analyze(
                    self.problem, param_samples, y[:, i], conf_level=conf_level
                )
            results[name] = Si  # type: ignore PGH003

        if method == "sobol":
            return _sobol_results_to_df(results)
        return _morris_results_to_df(results, self.problem)

    @staticmethod
    def plot_sobol(
        results: pd.DataFrame,
        index: str = "S1",
        n_cols: int | None = None,
        figsize: tuple | None = None,
    ):
        """
        Plot Sobol sensitivity analysis results.

        Parameters
        ----------
        results: pd.DataFrame
            The results from sobol_results_to_df.
        index: str
            The type of sensitivity index to plot.
            - "S1": first-order indices
            - "S2": second-order/interaction indices
            - "ST": total-order indices
            Defaults to "S1".
        n_cols: int | None
            The number of columns in the plot. Defaults to 3 if there are 3 or
            more outputs, otherwise the number of outputs. Defaults to None.
        figsize: tuple | None
            Figure size as (width, height) in inches. If None, set automatically.
            Defaults to None.
        """
        return _plot_sobol_analysis(results, index, n_cols, figsize)

    @staticmethod
    def plot_morris(
        results: pd.DataFrame,
        param_groups: dict[str, list[str]] | None = None,
        n_cols: int | None = None,
        figsize: tuple | None = None,
    ):
        """
        Plot Morris analysis results.

        Parameters
        ----------
        results: pd.DataFrame
            The results from sobol_results_to_df.
        param_groups: dic[str, list[str]] | None
            Optional parameter groupings used to give all the same plot color
            of the form ({<group name> : [param1, ...], }).
        n_cols: int, optional
            The number of columns in the plot. Defaults to 3 if there are 3 or
            more outputs, otherwise the number of outputs.
        figsize: tuple, optional
            Figure size as (width, height) in inches.If None, set calculated.
        """
        return _plot_morris_analysis(results, param_groups, n_cols, figsize)

    @staticmethod
    def top_n_sobol_params(
        sa_results_df: pd.DataFrame, top_n: int, sa_index: str = "ST"
    ) -> list:
        """
        Return `top_n` most important parameters given Sobol sensitivity analysis.

        In case of multiple outputs, averages over them to rank the parameters.

        Parameters
        ----------
        sa_results_df: pd.DataFrame
            Dataframe results by `SensitivityAnalysis().run()`
        top_n: int
            Number of parameters to return.
        sa_index: str
            Which sensitivity index to rank the parameters by. One of ["S1", "S2", "ST].

        Returns
        -------
        list[str]
            List of `top_n` parameter names.
        """
        if not all(
            col in sa_results_df.columns for col in ["index", "parameter", "value"]
        ):
            msg = (
                "sa_results_df is missing required columns: 'index', 'parameter',"
                "or 'value'"
            )
            raise ValueError(msg)

        st_results = sa_results_df[sa_results_df["index"] == sa_index]

        return (
            st_results.groupby("parameter")["value"]  # pyright: ignore[reportCallIssue]
            # each parameter is evalued against each output
            # to rank parameters, average over how sensitive all outputs are to it
            .mean()
            .nlargest(top_n)
            .index.tolist()
        )


def _sobol_results_to_df(results: dict[str, ResultDict]) -> pd.DataFrame:
    """
    Convert Sobol results to a (long-format) pandas DataFrame.

    Parameters
    ----------
    results: dict
        The Sobol indices returned by sobol_analysis.

    Returns
    -------
    pd.DataFrame
        A DataFrame with columns: 'output', 'parameter', 'index', 'value', 'confidence'.
    """
    rename_dict = {
        "variable": "index",
        "S1": "value",
        "S1_conf": "confidence",
        "ST": "value",
        "ST_conf": "confidence",
        "S2": "value",
        "S2_conf": "confidence",
    }
    rows = []
    for output, result in results.items():
        s1, st, s2 = result.to_df()
        s1 = (
            s1.reset_index()
            .rename(columns={"index": "parameter"})
            .rename(columns=rename_dict)
        )
        s1["index"] = "S1"
        st = (
            st.reset_index()
            .rename(columns={"index": "parameter"})
            .rename(columns=rename_dict)
        )
        st["index"] = "ST"
        s2 = (
            s2.reset_index()
            .rename(columns={"index": "parameter"})
            .rename(columns=rename_dict)
        )
        s2["index"] = "S2"

        df = pd.concat([s1, st, s2])
        df["output"] = output
        rows.append(df[["output", "parameter", "index", "value", "confidence"]])

    return pd.concat(rows)


def _validate_input(results: pd.DataFrame, index: str):
    if not isinstance(results, pd.DataFrame):
        results = _sobol_results_to_df(results)
        # we only want to plot one index type at a time
    valid_indices = ["S1", "S2", "ST"]
    if index not in valid_indices:
        raise ValueError(
            f"Invalid index type: {index}. Must be one of {valid_indices}."
        )
    return results[results["index"].isin([index])]


def _calculate_layout(n_outputs: int, n_cols: int | None = None):
    """Calculate plot layout (n rows, n cols)."""
    if n_cols is None:
        n_cols = 3 if n_outputs >= 3 else n_outputs
    n_rows = int(np.ceil(n_outputs / n_cols))
    return n_rows, n_cols


def _create_bar_plot(ax: Axes, output_data: pd.DataFrame, output_name: str):
    """Create a bar plot for a single output."""
    bar_color = "#4C4B63"
    x_pos = np.arange(len(output_data))
    conf = output_data["confidence"].values
    assert isinstance(conf, NumpyLike)
    yerr = conf / 2

    _ = ax.bar(
        x_pos,
        output_data["value"],
        color=bar_color,
        yerr=yerr,
        capsize=3,
    )

    ax.set_xticks(x_pos)
    ax.set_xticklabels(output_data["parameter"], rotation=45, ha="right")
    ax.set_ylabel("Sobol Index")
    ax.set_title(f"Output: {output_name}")


def _plot_sobol_analysis(
    results: pd.DataFrame,
    index: str = "S1",
    n_cols: int | None = None,
    figsize: tuple | None = None,
):
    """
    Plot the sobol sensitivity analysis results.

    Parameters
    ----------
    results: pd.DataFrame
        The results from sobol_results_to_df.
    index: str, default "S1"
        The type of sensitivity index to plot.
        - "S1": first-order indices
        - "S2": second-order/interaction indices
        - "ST": total-order indices
    n_cols: int | None
        The number of columns in the plot. If None, sets n_cols to 3 if there are 3 or
        more outputs, otherwise the number of outputs. Defaults to None.
    figsize: tuple | None
        Figure size as (width, height) in inches.If None, automatically calculated.
        Defaults to None.
    """
    with plt.style.context("fast"):
        # prepare data
        results = _validate_input(results, index)  # pyright: ignore[reportAssignmentType]
        unique_outputs = results["output"].unique()
        n_outputs = len(unique_outputs)

        # layout
        n_rows, n_cols = _calculate_layout(n_outputs, n_cols)
        figsize = figsize or (4.5 * n_cols, 4 * n_rows)

        fig, axes = plt.subplots(n_rows, n_cols, figsize=figsize)
        if isinstance(axes, np.ndarray):
            axes = axes.flatten()
        elif n_outputs == 1:
            axes = [axes]

        for ax, output in zip(axes, unique_outputs, strict=False):
            output_data = results[results["output"] == output]
            assert isinstance(output_data, pd.DataFrame)
            _create_bar_plot(ax, output_data, output)

        # remove any empty subplots
        for idx in range(len(unique_outputs), len(axes)):
            fig.delaxes(axes[idx])

        index_names = {
            "S1": "First-Order",
            "S2": "Second-order/Interaction",
            "ST": "Total-Order",
        }

        # title
        fig.suptitle(
            f"{index_names[index]} indices and 95% CI",
            fontsize=14,
        )

        plt.tight_layout()

    return display_figure(fig)


def _morris_results_to_df(
    results: dict[str, ResultDict], problem: dict
) -> pd.DataFrame:
    """
    Convert Morris results to a (long-format) pandas DataFrame.

    Parameters
    ----------
    results: dict
        The Morris indices returned by morris_analysis.
    problem: dict
        The problem definition, including 'names'.

    Returns
    -------
    pd.DataFrame
        A DataFrame with columns: 'output', 'parameter', 'mu', 'mu_star', 'sigma',
        'mu_star_conf'.
    """
    rows = []
    parameter_names = problem["names"]

    for output, result in results.items():
        df_data = {
            "output": [output] * len(parameter_names),
            "parameter": parameter_names,
            "mu": result["mu"],
            "mu_star": result["mu_star"],
            "sigma": result["sigma"],
            "mu_star_conf": result["mu_star_conf"],
        }

        df = pd.DataFrame(df_data)
        rows.append(df)

    return pd.concat(rows, ignore_index=True)


def _plot_morris_analysis(
    results: pd.DataFrame,
    param_groups: dict[str, list[str]] | None = None,
    n_cols: int | None = None,
    figsize: tuple | None = None,
):
    """
    Plot the Morris sensitivity analysis results.

    Parameters
    ----------
    results: pd.DataFrame
        The results from morris_results_to_df.
    param_groups: dict, optional
        Dictionary mapping parameter names to groups for coloring.
    n_cols: int, optional
        The number of columns in the plot. Defaults to 3 if there are 3 or more outputs,
        otherwise the number of outputs.
    figsize: tuple, optional
        Figure size as (width, height) in inches. If None, automatically calculated.
    """
    with plt.style.context("fast"):
        unique_outputs = results["output"].unique()
        n_outputs = len(unique_outputs)

        # layout - add space for legend
        n_rows, n_cols = _calculate_layout(n_outputs, n_cols)
        figsize = figsize or (4.5 * n_cols + 2, 4 * n_rows)  # Extra width for legend

        fig, axes = plt.subplots(n_rows, n_cols, figsize=figsize)
        if isinstance(axes, np.ndarray):
            axes = axes.flatten()
        elif n_outputs == 1:
            axes = [axes]

        # Create color mappings once for all plots
        colors = [
            "#4C4B63",
            "#E63946",
            "#F77F00",
            "#FCBF49",
            "#06D6A0",
            "#118AB2",
            "#073B4C",
            "#FF6B6B",
            "#4ECDC4",
            "#45B7D1",
            "#96CEB4",
            "#FFEAA7",
            "#DDA0DD",
            "#98D8C8",
        ]

        if param_groups is None:
            # Different color for each parameter
            all_params = results["parameter"].unique()
            param_colors = {
                param: colors[i % len(colors)] for i, param in enumerate(all_params)
            }
            legend_items = [(param, param_colors[param]) for param in all_params]
            legend_title = "Parameters"
        else:
            # Color by parameter groups
            unique_groups = list(set(param_groups.values()))
            param_colors = {
                group: colors[i % len(colors)] for i, group in enumerate(unique_groups)
            }
            legend_items = [(group, param_colors[group]) for group in unique_groups]
            legend_title = "Parameter Groups"

        # Plot each output
        for ax, output in zip(axes, unique_outputs, strict=False):
            output_data = results[results["output"] == output]
            assert isinstance(output_data, pd.DataFrame)
            _create_morris_plot(
                ax,
                output_data,
                output,
                param_groups,
                param_colors,
            )

        # remove any empty subplots
        for idx in range(len(unique_outputs), len(axes)):
            fig.delaxes(axes[idx])

        # Create single legend on the right side
        legend_handles = []
        legend_labels = []
        for label, color in legend_items:
            handle = Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                markerfacecolor=color,
                markersize=8,
                alpha=0.7,
                linewidth=0,
            )
            legend_handles.append(handle)
            legend_labels.append(label)

        # Add legend to the right of the plots
        fig.legend(
            legend_handles,
            legend_labels,
            loc="center right",
            bbox_to_anchor=(0.98, 0.5),
            title=legend_title,
            framealpha=0.9,
            fontsize=10,
        )

        # title
        fig.suptitle(
            r"Morris Sensitivity Analysis ($\mu^*$ vs $\sigma$)",
            fontsize=14,
        )

        plt.tight_layout()

    return display_figure(fig)


def _create_morris_plot(
    ax: Axes,
    output_data: pd.DataFrame,
    output_name: str,
    param_groups: dict | None = None,
    color_mapping: dict | None = None,
):
    """Create a Morris plot (mu_star vs sigma) for a single output."""
    # Default colors - expanded palette for more variety
    colors = [
        "#4C4B63",
        "#E63946",
        "#F77F00",
        "#FCBF49",
        "#06D6A0",
        "#118AB2",
        "#073B4C",
        "#FF6B6B",
        "#4ECDC4",
        "#45B7D1",
        "#96CEB4",
        "#FFEAA7",
        "#DDA0DD",
        "#98D8C8",
    ]

    # Use provided color mapping or create default
    if color_mapping is None:
        if param_groups is None:
            # Different color for each parameter when no groups specified
            unique_params = output_data["parameter"].unique()
            param_colors = {
                param: colors[i % len(colors)] for i, param in enumerate(unique_params)
            }
            color_mapping = param_colors
        else:
            # Color by parameter groups
            unique_groups = list(set(param_groups.values()))
            group_colors = {
                group: colors[i % len(colors)] for i, group in enumerate(unique_groups)
            }
            color_mapping = group_colors

    # Plot points without labels for legend (legend is handled at figure level)
    for _, row in output_data.iterrows():
        param_name = row["parameter"]

        if param_groups is None:
            color = color_mapping[param_name]
        else:
            group = param_groups.get(param_name, "default")
            color = color_mapping.get(group, colors[0])

        ax.scatter(row["sigma"], row["mu_star"], color=color, alpha=0.7, s=60)

    # Add parameter labels with matching colors
    for _, row in output_data.iterrows():
        param_name = row["parameter"]
        assert isinstance(param_name, str)

        if param_groups is None:
            # Use same color as the dot (individual parameter color)
            label_color = color_mapping[param_name]
        else:
            # Use group color
            group = param_groups.get(param_name, "default")
            label_color = color_mapping.get(group, colors[0])

        x_coord, y_coord = row["sigma"], row["mu_star"]
        assert isinstance(x_coord, float)
        assert isinstance(y_coord, float)
        ax.annotate(
            param_name,
            (x_coord, y_coord),
            xytext=(5, 5),
            textcoords="offset points",
            fontsize=8,
            alpha=0.9,
            color=label_color,
            fontweight="bold",
        )

    ax.set_xlabel("σ (Standard Deviation)")  # noqa: RUF001
    ax.set_ylabel("μ* (Modified Mean)")
    ax.set_title(f"Output: {output_name}")
    ax.grid(True, alpha=0.3)



import logging
from abc import ABC, abstractmethod

import torch
from tqdm import tqdm

from autoemulate.core.logging_config import get_configured_logger
from autoemulate.core.types import TensorLike
from autoemulate.data.utils import ValidationMixin, set_random_seed
from autoemulate.simulations.experimental_design import LatinHypercube

logger = logging.getLogger("autoemulate")


class Simulator(ABC, ValidationMixin):
    """
    Base class for simulations. All simulators should inherit from this class.

    This class provides the interface and common functionality for different
    simulation implementations.
    """

    def __init__(
        self,
        parameters_range: dict[str, tuple[float, float]],
        output_names: list[str],
        log_level: str = "progress_bar",
    ):
        """
        Initialize the simulator with parameter ranges and output names.

        Parameters
        ----------
        parameters_range: dict[str, tuple[float, float]]
            Dictionary mapping input parameter names to their (min, max) ranges.
        output_names: list[str]
            List of output parameters' names.
        log_level: str
            Logging level for the simulator. Can be one of:
            - "progress_bar": shows a progress bar during batch simulations
            - "debug": shows debug messages
            - "info": shows informational messages
            - "warning": shows warning messages
            - "error": shows error messages
            - "critical": shows critical messages
        """
        self._parameters_range = parameters_range
        self._param_names = list(parameters_range.keys())
        self._param_bounds = list(parameters_range.values())
        self._output_names = output_names
        self._in_dim = len(self.param_names)
        self._out_dim = len(self.output_names)
        self._has_sample_forward = False
        self.logger, self.progress_bar = get_configured_logger(log_level)

    @classmethod
    def simulator_name(cls) -> str:
        """Get the name of the simulator class."""
        return cls.__name__

    @property
    def parameters_range(self) -> dict[str, tuple[float, float]]:
        """Dictionary mapping input parameter names to their (min, max) ranges."""
        return self._parameters_range

    @parameters_range.setter
    def parameters_range(
        self, parameters_range: dict[str, tuple[float, float]]
    ) -> None:
        """Set the range of input parameters for the simulator.

        Parameters
        ----------
        parameters_range: dict[str, tuple[float, float]]
            Dictionary mapping input parameter names to their (min, max) ranges.
        """
        self._parameters_range = parameters_range
        self._param_names = list(parameters_range.keys())
        self._param_bounds = list(parameters_range.values())
        self._in_dim = len(self.param_names)

    @property
    def param_names(self) -> list[str]:
        """List of parameter names."""
        return self._param_names

    @property
    def param_bounds(self) -> list[tuple[float, float]]:
        """List of parameter bounds."""
        return self._param_bounds

    @property
    def output_names(self) -> list[str]:
        """List of output parameter names."""
        return self._output_names

    @output_names.setter
    def output_names(self, output_names: list[str]) -> None:
        """Set the names of output parameters for the simulator.

        This setter allows renaming the output parameters but does not allow
        changing the number of outputs (dimensionality is fixed after initialization).

        Parameters
        ----------
        output_names: list[str]
            List of output parameter names. Must have the same length as the current
            number of outputs.

        Raises
        ------
        ValueError
            If the number of output names differs from the simulator's fixed output
            dimension.
        """
        if len(output_names) != self._out_dim:
            raise ValueError(
                f"Number of output names ({len(output_names)}) must match "
                f"simulator output dimension ({self._out_dim}). Cannot change "
                f"dimensionality after initialization."
            )
        self._output_names = output_names

    @property
    def in_dim(self) -> int:
        """Input dimensionality."""
        return self._in_dim

    @property
    def out_dim(self) -> int:
        """Output dimensionality."""
        return self._out_dim

    def sample_inputs(
        self, n_samples: int, random_seed: int | None = None
    ) -> TensorLike:
        """
        Generate random samples using Latin Hypercube Sampling.

        Parameters
        ----------
        n_samples: int
            Number of samples to generate.
        random_seed: int | None
            Random seed for reproducibility. If None, no seed is set.

        Returns
        -------
        TensorLike
            Parameter samples (column order is given by self.param_names)
        """
        if random_seed is not None:
            set_random_seed(random_seed)  # type: ignore PGH003
        lhd = LatinHypercube(self.param_bounds)
        return lhd.sample(n_samples)

    @abstractmethod
    def _forward(self, x: TensorLike) -> TensorLike | None:
        """
        Abstract method to perform the forward simulation.

        Parameters
        ----------
        x: TensorLike
            Input parameters into the simulation forward run.

        Returns
        -------
        TensorLike | None
            Simulated output tensor. Shape = (1, self.out_dim).
            For example, if the simulator outputs two simulated variables,
            then the shape would be (1, 2). None if the simulation fails.
        """

    def forward(self, x: TensorLike) -> TensorLike | None:
        """
        Generate samples from input data using the simulator.

        Combines the abstract method `_forward` with some validation checks.

        Parameters
        ----------
        x: TensorLike
            Input tensor of shape (n_samples, self.in_dim).

        Returns
        -------
        TensorLike
            Simulated output tensor. None if the simulation failed.
        """
        y = self._forward(self.check_matrix(x))
        if isinstance(y, TensorLike):
            y = self.check_matrix(y)
            x, y = self.check_pair(x, y)
            return y
        return None

    def forward_batch(self, x: TensorLike) -> TensorLike:
        """Run multiple simulations with different parameters.

        For infallible simulators that always succeed.
        If your simulator might fail, use `forward_batch_skip_failures()` instead.

        Parameters
        ----------
        x: TensorLike
            Tensor of input parameters to make predictions for.

        Returns
        -------
        TensorLike
            Tensor of simulation results of shape (n_batch, self.out_dim).

        Raises
        ------
        RuntimeError
            If the number of simulations does not match the input.
            Use `forward_batch_skip_failures()` to handle failures.
        """
        results, x_valid = self.forward_batch_skip_failures(x)

        # Raise an error if the number of simulations does not match the input
        if x.shape[0] != x_valid.shape[0]:
            msg = (
                "Some simulations failed. Use forward_batch_skip_failures() to handle "
                "failures."
            )
            raise RuntimeError(msg)

        return results

    def forward_batch_skip_failures(
        self, x: TensorLike
    ) -> tuple[TensorLike, TensorLike]:
        """Run multiple simulations, skipping any that fail.

        For simulators where for some inputs the simulation can fail.
        Failed simulations are skipped, and only successful results are returned
        along with their corresponding input parameters.

        Parameters
        ----------
        x: TensorLike
            Tensor of input parameters to make predictions for.

        Returns
        -------
        tuple[TensorLike, TensorLike]
            Tuple of (simulation_results, valid_input_parameters).
            Only successful simulations are included.
        """
        self.logger.info("Running batch simulation for %d samples", len(x))

        results = []
        successful = 0
        valid_idx = []

        # Process each sample with progress tracking
        for i in tqdm(
            range(len(x)),
            desc="Running simulations",
            disable=not self.progress_bar,
            total=len(x),
            unit="sample",
            unit_scale=True,
        ):
            logger.debug("Running simulation for sample %d/%d", i + 1, len(x))
            result = self.forward(x[i : i + 1])
            if result is not None:
                results.append(result)
                successful += 1
                valid_idx.append(i)
                logger.debug("Simulation %d/%d successful", i + 1, len(x))
            else:
                logger.warning(
                    "Simulation %d/%d failed. Result is None.", i + 1, len(x)
                )

        # Report results
        self.logger.info(
            "Successfully completed %d/%d simulations (%.1f%%)",
            successful,
            len(x),
            (successful / len(x) * 100 if len(x) > 0 else 0.0),
        )

        # stack results into a 2D array on first dim using torch
        results_tensor = torch.cat(results, dim=0)

        return results_tensor, x[valid_idx]

    def get_parameter_idx(self, name: str) -> int:
        """
        Get the index of a specific parameter.

        Parameters
        ----------
        name: str
            Name of the parameter to retrieve.

        Returns
        -------
        float
            Index of the specified parameter.
        """
        if name not in self._param_names:
            raise ValueError(f"Parameter {name} not found.")
        return self._param_names.index(name)



from .epidemic import Epidemic
from .flow_problem import FlowProblem
from .projectile import Projectile, ProjectileMultioutput

ALL_SIMULATORS = [Epidemic, FlowProblem, Projectile, ProjectileMultioutput]

__all__ = ["Epidemic", "FlowProblem", "Projectile", "ProjectileMultioutput"]

SIMULATOR_REGISTRY = dict(zip(__all__, ALL_SIMULATORS, strict=False))



from typing import Protocol

import torch
from gpytorch.kernels import Kernel
from gpytorch.means import Mean


class MeanModuleFn(Protocol):
    """Protocol for mean module functions."""

    def __call__(self, n_features: int | None, n_outputs: torch.Size | None) -> Mean:
        """Call the mean module function with the specified parameters."""
        ...


class CovarModuleFn(Protocol):
    """Protocol for covariance module functions."""

    # TODO: consider revising input API to be more flexible
    def __call__(self, n_features: int | None, n_outputs: torch.Size | None) -> Kernel:
        """Call the covariance module function with the specified parameters."""
        ...



from typing import cast

from torch.distributions import ComposeTransform, Transform, TransformedDistribution

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import (
    DeviceLike,
    DistributionLike,
    GaussianLike,
    GaussianProcessLike,
    OutputLike,
    TensorLike,
)
from autoemulate.data.utils import ValidationMixin
from autoemulate.emulators.base import Emulator
from autoemulate.transforms.base import (
    AutoEmulateTransform,
    _inverse_sample_gaussian_like,
    _inverse_sample_gaussian_process_like,
)


class TransformedEmulator(Emulator, ValidationMixin):
    """
    A transformed emulator that applies transformations to input and target data.

    This class wraps an emulator model with configurable transformations applied to
    both input features (x) and target variables (y). The emulator is trained and
    makes predictions in the transformed space, with automatic inverse transformations
    applied to return results in the original data space.

    The transformation workflow for fitting:
    ```
    Original space:     x                                         y
                        │                                         │
                        ▼ x_transforms.fit()   y_transforms.fit() ▼
                        │                                         │
    Transformed space:  x_t ──────────► emulator.fit() ◄──────── y_t
    ```


    The transformation workflow for prediction:
    ```
    Original space:     x ────────────────────► y_pred
                        │                       ▲
                        │ x_transforms          │ y_transforms⁻¹
                        ▼                       │
    Transformed space:  x_t ──► emulator ──► y_t_pred
    ```

    Key features:
    - Sequential application of multiple transformations
    - Automatic handling of different prediction output types (tensors, distributions)
    - Support for both analytical and sampling-based inverse transformations
    - Configurable behavior for high-dimensional targets

    Attributes
    ----------
    x_transforms: list[AutoEmulateTransform]
        List of transformations applied to input data (x) in sequential order.
    model: Emulator
        The underlying emulator model that operates on transformed data.
    y_transforms: list[AutoEmulateTransform]
        List of transformations applied to target data (y) in sequential order.
    """

    x_transforms: list[AutoEmulateTransform]
    model: Emulator
    y_transforms: list[AutoEmulateTransform]

    def __init__(  # noqa: PLR0913
        self,
        x: TensorLike,
        y: TensorLike,
        x_transforms: list[AutoEmulateTransform] | None,
        y_transforms: list[AutoEmulateTransform] | None,
        model: type[Emulator],
        output_from_samples: bool = False,
        n_samples: int = 1000,
        full_covariance: bool = False,
        max_targets: int = 200,
        device: DeviceLike | None = None,
        **kwargs,
    ):
        """
        Initialize a transformed emulator.

        Parameters
        ----------
        x: TensorLike
            Input training data tensor of shape (n_samples, n_features).
        y: TensorLike
            Target training data tensor of shape (n_samples, n_targets).
        x_transforms: list[AutoEmulateTransform] | None
            List of transforms to apply to input data in sequential order.
            If None, no transformations are applied to x.
        y_transforms: list[AutoEmulateTransform] | None
            List of transforms to apply to target data in sequential order.
            If None, no transformations are applied to y.
        model: type[Emulator]
            The emulator class to instantiate and train on transformed data.
        output_from_samples: bool
            Whether to obtain predictions by sampling from the model's predictive
            distribution. Automatically set to True for high-dimensional targets
            (n_targets > max_targets). Defaults to False.
        n_samples: int
            Number of samples to draw when using sampling-based predictions.
            Only used when output_from_samples=True. Defauls to 1000.
        full_covariance: bool
            Whether to use full covariance matrix for predictions. If False,
            uses diagonal covariance. Automatically set to False for
            high-dimensional targets (n_targets > max_targets). Defaults to False.
        max_targets: int
            Threshold for switching to approximate sampling-based predictions
            with diagonal covariance when dealing with high-dimensional targets.
            Defaults to 200.
        device: DeviceLike | None
            Device for tensor operations. If None, uses the default device.
            Defaults to None.
        **kwargs
            Additional keyword arguments passed to the emulator constructor.

        Notes
        -----
        - Transforms are fitted on the provided training data during initialization
        - The underlying emulator is trained on the transformed data
        - For targets with dimensionality > max_targets, the emulator automatically
          switches to sampling-based predictions with diagonal covariance for efficiency
        """
        self.x_transforms = x_transforms or []
        self.y_transforms = y_transforms or []
        self._fit_transforms(x, y)
        self.untransformed_model_name = model.model_name()
        self.model = model(
            self._transform_x(x), self._transform_y_tensor(y), device=device, **kwargs
        )
        self.output_from_samples = output_from_samples or y.shape[1] > max_targets
        self.n_samples = n_samples
        self.full_covariance = full_covariance and y.shape[1] <= max_targets
        TorchDeviceMixin.__init__(self, device=device)
        self.supports_grad = self.model.supports_grad and all(
            t.bijective for t in self.x_transforms
        )

    def _fit_transforms(self, x: TensorLike, y: TensorLike):
        """
        Fit the transforms on the provided training data.

        Parameters
        ----------
        x: TensorLike
            Input training data tensor of shape `(n_samples, n_features)`.
        y: TensorLike
            Target training data tensor of shape `(n_samples, n_targets)`.
        """
        # Fit transforms
        current_x = x
        for transform in self.x_transforms:
            transform.fit(current_x)
            current_x = transform(current_x)
        # Fit target transforms
        current_y = y
        for transform in self.y_transforms:
            transform.fit(current_y)
            current_y = transform(current_y)

    def refit(self, x: TensorLike, y: TensorLike, retrain_transforms: bool = False):
        """
        Refit the emulator with new data and optionally retrain transforms.

        Parameters
        ----------
        x: TensorLike
            New input training data tensor of shape `(n_samples, n_features)`.
        y: TensorLike
            New target training data tensor of shape `(n_samples, n_targets)`.
        retrain_transforms: bool
            Whether to retrain the transforms on the new data. If False,
            uses the existing fitted transforms from initialization. Defaults to False.

        Notes
        -----
        When retrain_transforms=False, the transforms fitted during initialization
        are applied to the new data. This assumes the new data comes from the same
        distribution as the original training data.
        """
        if retrain_transforms:
            self._fit_transforms(x, y)
        self.fit(x, y)

    def _transform_x(self, x: TensorLike) -> TensorLike:
        """
        Transform the input tensor `x` using `x_transforms` returning a `TensorLike`.

        Parameters
        ----------
        x: TensorLike
            Input tensor to be transformed.

        Returns
        -------
        TensorLike
            Transformed input tensor after applying all x_transforms.

        """
        x_t = ComposeTransform(self._cast(self.x_transforms))(x)
        assert isinstance(x_t, TensorLike)
        return x_t

    def _transform_y_tensor(self, y: TensorLike) -> TensorLike:
        """
        Transform the target tensor `y` using `y_transforms` returning a `TensorLike`.

        Parameters
        ----------
        y: TensorLike
            Target tensor to be transformed.

        Returns
        -------
        TensorLike
            Transformed target tensor after applying all `y_transforms`.

        """
        y_t = ComposeTransform(self._cast(self.y_transforms))(y)
        assert isinstance(y_t, TensorLike)
        return y_t

    @staticmethod
    def _cast(transforms: list[AutoEmulateTransform]) -> list[Transform]:
        """Casts a list of AutoEmulateTransform to a list of torch Transforms."""
        return cast(list[Transform], transforms)

    def _inv_transform_y_tensor(self, y_t: TensorLike) -> TensorLike:
        """
        Invert the transformed target tensor `y_t` back to the original space.

        Parameters
        ----------
        y_t: TensorLike
            Transformed target tensor to be inverted.

        Returns
        -------
        TensorLike
            Inverted target tensor in the original data space after applying all
            inverse `y_transforms`.
        """
        target_transforms = self._cast(self.y_transforms)
        y = ComposeTransform(target_transforms).inv(y_t)
        assert isinstance(y, TensorLike)
        return y

    def _inv_transform_y_gaussian(self, y_t: GaussianLike) -> GaussianLike:
        """
        Invert the transformed `GaussianLike` `y_t` back to the original space.

        The inversion is performed with calls to each transform's inverse_gaussian
        method that aims to use the analytical or approximate non-sampling inverse of
        the transformation for Gaussian distributions.

        Parameters
        ----------
        y_t: GaussianLike
            Transformed MultitaskMultivariateNormal target distribution to be inverted.

        Returns
        -------
        GaussianLike
            Inverted GaussianLike distribution in the original data space after applying
            all inverse `y_transforms` with the transforms `inverse_gaussian` methods.
        """
        # Invert the order since the combined transform is an inversion
        for transform in self.y_transforms[::-1]:
            y_t = transform._inverse_gaussian(y_t)
        return y_t

    def _inv_transform_y_gaussian_sample(
        self, y_t: DistributionLike
    ) -> GaussianLike | GaussianProcessLike:
        """
        Invert the transformed distribution `y_t` by sampling.

        Invert the transformed distribution `y_t` by sampling and calculating
        empirical mean and covariance from the samples in the original space to
        parameterize a `GaussianLike` distribution.

        This method accepts any `DistributionLike` input but returns `GaussianLike` or
        `GaussianProcessLike` distributions.

        This method uses the number of samples specified in the initialization
        (`n_samples`) to draw samples from the transformed distribution `y_t` and
        returns a full covariance `GaussianLike` if specified (`full_covariance=True`
        in the initialization and fewer than `max_targets`) or a diagonal covariance
        `GaussianLike` otherwise for computational feasibility.

        Parameters
        ----------
        y_t: DistributionLike
            Transformed target distribution to be inverted by sampling.

        Returns
        -------
        GaussianLike | GaussianProcessLike
            A `GaussianProcessLike` distribution if the input was `GaussianProcessLike`,
            or a `GaussianLike` distribution if the input was any other
            `DistributionLike`. The distribution is parameterized by the empirical mean
            and covariance of the samples drawn from the transformed distribution in the
            original data space after applying all inverse `y_transforms`.

        Raises
        ------
        RuntimeError
            If the empirical covariance cannot be made positive definite.

        Notes
        -----
        This method can be used when the emulator's predictive distribution is not
        `GaussianLike` or when analytical or alternative approximate inversion is not
        possible.
        """
        # Handle GaussianProcessLike distinctly
        if isinstance(y_t, GaussianProcessLike):
            return _inverse_sample_gaussian_process_like(
                self._inv_transform_y_tensor, y_t, self.n_samples, self.full_covariance
            )

        # If `y_t` is not a `GaussianProcessLike`, sample from it and return a
        # `GaussianLike`
        return _inverse_sample_gaussian_like(
            self._inv_transform_y_tensor, y_t, self.n_samples, self.full_covariance
        )

    def _inv_transform_y_distribution(self, y_t: DistributionLike) -> DistributionLike:
        """
        Invert the transformed distribution `y_t` back to the original space `y`.

        This method applies the inverse transformations to the distribution `y_t`
        using the `inv` method of the `ComposeTransform` class, which is a composition
        of all inverse transforms in `y_transforms`.

        Parameters
        ----------
        y_t: DistributionLike

        Returns
        -------
        DistributionLike
            Inverted distribution `y` in the original data space after applying all
            inverse `y_transforms`.

        Raises
        ------
        RuntimeError
            If the distribution cannot be inverted due to non-bijective transforms.

        Notes
        -----
        The method requires that all transforms in `y_transforms` are bijective
        with log det Jacobian defined so that the returned transformed distribution is
        valid. As such, this method is:
        - is not used in dimensionality reduction contexts (e.g. with PCA, VAE, etc.)
        - does not return a distribution with mean and variance readily implemented
          (this would require further empirical estimation from samples from the
          returned transformed distribution `y`).
        """
        target_transforms = self._cast(self.y_transforms)
        return TransformedDistribution(y_t, [ComposeTransform(target_transforms).inv])

    def _fit(self, x: TensorLike, y: TensorLike):
        # Transform x and y
        x_t = self._transform_x(x)
        y_t = self._transform_y_tensor(y)

        # Fit on transformed variables
        self.model.fit(x_t, y_t)

    def _predict(self, x: TensorLike, with_grad: bool) -> OutputLike:
        if with_grad and not self.supports_grad:
            msg = "Gradient calculation is not supported."
            raise ValueError(msg)

        # Transform and invert transform for prediction in original data space
        x_t = self._transform_x(x)
        y_t_pred = self.model.predict(x_t, with_grad)

        # If TensorLike, transform tensor back to original space
        if isinstance(y_t_pred, TensorLike):
            return self._inv_transform_y_tensor(y_t_pred)

        # Output derived by analytical/approximate transformations
        if not self.output_from_samples:
            if isinstance(y_t_pred, GaussianLike):
                return self._inv_transform_y_gaussian(y_t_pred)
            if isinstance(y_t_pred, DistributionLike):
                return self._inv_transform_y_distribution(y_t_pred)
            msg = "y_pred is not TensorLike, GaussianLike or DistributionLike"
            raise ValueError(msg)

        # Output derived by sampling and inverting to original space
        if isinstance(y_t_pred, DistributionLike):
            return self._inv_transform_y_gaussian_sample(y_t_pred)
        msg = (
            "Invalid output type from model prediction. Expected TensorLike,"
            "GaussianLike, or DistributionLike. Received: "
            f"{type(y_t_pred)}"
        )
        raise ValueError(msg)

    @staticmethod
    def is_multioutput() -> bool:
        """Not implemented for TransformedEmulator.

        TransformedEmulator does not implement is_multioutput as a staticmethod
        since it depends on the emulator instance.
        """
        msg = (
            "TransformedEmulator does not implement is_multioutput as a staticmethod"
            "since it depends on the emulator instance."
        )
        raise NotImplementedError(msg)


# TODO: implement TransformedModuleEmulator with learnable parameters
# class TransformedModuleEmulator(Emulator, nn.Module, ValidationMixin):
#     def __init__(
#         self,
#         x: TensorLike,
#         y: TensorLike,
#         transforms: list[AutoEmulateTransformModule],
#         target_transform: list[AutoEmulateTransformModule],
#         model: nn.Module,
#     ):
#         _, _ = x, y
#         self.transforms = transforms
#         self.target_transform = target_transform
#         self.model = model

#     def forward(self, x):
#         x = self._transform_x(x)
#         return self._inv_transform_y_tensor(self.model(x))

#     # TODO: fix the types here
#     def _transform_x(self, x: TensorLike) -> TensorLike:
#         return ComposeTransformModule(self.transforms)(x)

#     def _transform_y_tensor(self, y: TensorLike) -> TensorLike:
#         return ComposeTransformModule(self.target_transforms)(y)

#     def _inv_transform_y_tensor(self, y: TensorLike) -> TensorLike:
#         return ComposeTransformModule(self.target_transforms).inv(y)

#     def _inv_transform_y_distribution(
#         self, y_dis: DistributionLike
#     ) -> DistributionLike:
#         return TransformedDistribution(
#             y_dis, [ComposeTransform(self.target_transform).inv]
#         )

#     def _fit(self, x, y): ...



# from https://github.com/alan-turing-institute/mogp-emulator/blob/main/mogp_emulator/demos/projectile.py
import numpy as np
import torch
from scipy.integrate import solve_ivp

from autoemulate.core.types import NumpyLike, TensorLike
from autoemulate.simulations.base import Simulator


class Projectile(Simulator):
    """
    Simulator of projectile motion.

    A projectile is launched from an initial height of 2 meters at an  angle of 45
    degrees and falls under the influence of gravity and air resistance. Drag is
    proportional to the square of the velocity. We would like to determine the distance
    travelled by the projectile as a function of the drag coefficient and the launch
    velocity.
    """

    def __init__(
        self,
        parameters_range: dict[str, tuple[float, float]] | None = None,
        output_names: list[str] | None = None,
        log_level: str = "progress_bar",
    ):
        if parameters_range is None:
            parameters_range = {"c": (-5.0, 1.0), "v0": (0.0, 1000)}
        if output_names is None:
            output_names = ["distance"]
        super().__init__(parameters_range, output_names, log_level)

    def _forward(self, x: TensorLike) -> TensorLike:
        """
        Simulate the projectile motion and return the distance travelled.

        Parameters
        ----------
        x: TensorLike
            Dictionary of input parameter values to simulate:
            - `c`: the drag coefficient on a log scale
            - `v0`: velocity

        Returns
        -------
        TensorLike
            Distance travelled by projectile.
        """
        assert x.shape[0] == 1, (
            f"Simulator._forward expects a single input, got {x.shape[0]}"
        )
        y = simulate_projectile(x.cpu().numpy()[0])
        return torch.tensor([y]).view(-1, 1)


class ProjectileMultioutput(Simulator):
    """
    Multi-output simulator of projectile motion.

    Simulator of projectile motion that outputs both the distance travelled by the
    projectile and its velocity on impact.
    """

    def __init__(
        self,
        parameters_range: dict[str, tuple[float, float]] | None = None,
        output_names: list[str] | None = None,
        log_level: str = "progress_bar",
    ):
        if parameters_range is None:
            parameters_range = {"c": (-5.0, 1.0), "v0": (0.0, 1000)}
        if output_names is None:
            output_names = ["distance", "impact_velocity"]
        super().__init__(parameters_range, output_names, log_level)

    def _forward(self, x: TensorLike) -> TensorLike:
        """
        Simulate the projectile motion with multiple outputs.

        Simulate projectile motion and return the distance travelled and impact
        velocity.

        Parameters
        ----------
        x: TensorLike
            Dictionary of input parameter values to simulate:
            - `c`: the drag coefficient on a log scale
            - `v0`: velocity

        Returns
        -------
        TensorLike
            Distance travelled by projectile and impact velocity.
        """
        assert x.shape[0] == 1, (
            f"Simulator._forward expects a single input, got {x.shape[0]}"
        )
        y = simulate_projectile_multioutput(x.cpu().numpy()[0])
        return torch.tensor([y])


def f(t: float, y: NumpyLike, c: float):  # noqa: ARG001
    """
    Compute RHS of system of differential equations, returning vector derivative.

    Parameters
    ----------
    t: float
        Time variable (not used).
    y: array
        Array of dependent variables (vx, vy, x, y).
    c: float
        Drag coefficient (non-negative).
    """
    # check inputs and extract
    assert len(y) == 4
    assert c >= 0.0

    vx = y[0]
    vy = y[1]

    # calculate derivatives
    dydt = np.zeros(4)

    dydt[0] = -c * vx * np.sqrt(vx**2 + vy**2)
    dydt[1] = -9.8 - c * vy * np.sqrt(vx**2 + vy**2)
    dydt[2] = vx
    dydt[3] = vy

    return dydt


def event(t: float, y: NumpyLike, c: float) -> float:  # noqa: ARG001
    """
    Event to trigger end of integration. Stops when projectile hits ground.

    Parameters
    ----------
    t: float
        Time variable (not used).
    y: array
        Array of dependent variables (vx, vy, x, y).
    c: float
        Drag coefficient (non-negative).

    Returns
    -------
    float
        The height of the projectile.
    """
    assert len(y) == 4
    assert c >= 0.0

    return y[3]


# indicate this event terminates the simulation
event.terminal = True  # pyright: ignore[reportFunctionMemberAccess]


def simulator_base(x: NumpyLike):
    """
    Simulate ODE system for projectile motion with drag.

    Returns distance projectile travels.

    Parameters
    ----------
    x: NumpyLike
        Array of input parameters (c, v0).

    Returns
    -------
    results: scipy.integrate.OdeResult
        Results of ODE integration.
    """
    # unpack values

    assert len(x) == 2
    assert x[1] > 0.0

    c = 10.0 ** x[0]
    v0 = x[1]

    # set initial conditions

    y0 = np.zeros(4)

    y0[0] = v0 / np.sqrt(2.0)
    y0[1] = v0 / np.sqrt(2.0)
    y0[3] = 2.0

    # run simulation
    return solve_ivp(f, (0.0, 1.0e8), y0, events=event, args=(c,))


def simulate_projectile(x: NumpyLike) -> float:
    """
    Return the distance travelled by the projectile.

    Distance is obtained by solving the ODE system for projectile motion with drag.

    Parameters
    ----------
    x: NumpyLike
        Array of input parameters (c, v0).

    Returns
    -------
    distance: float
        Distance travelled by projectile.
    """
    results = simulator_base(x)

    return results.y_events[0][0][2]


def simulate_projectile_multioutput(x: NumpyLike) -> tuple[float, float]:
    """
    Return the distance travelled by the projectile and its impact velocity.

    Simulator to solve ODE system with multiple outputs.

    Parameters
    ----------
    x: NumpyLike
        Array of input parameters (c, v0).

    Returns
    -------
    float, float
        Distance travelled by projectile and its velocity on impact.
    """
    results = simulator_base(x)

    return (
        results.y_events[0][0][2],
        np.sqrt(results.y_events[0][0][0] ** 2 + results.y_events[0][0][1] ** 2),
    )



from autoemulate.learners.base import Emulator

from . import stream

# from .base import Emulator, Simulator
from .base import Simulator

# # TODO: automate this with Learner.hierarchy()
# # lowercase classes below should correspond to abstract classesß
# class active:
#     class stream:
#         Random = _stream.Random
#         class threshold:

#             # Operate on the input space X
#             class input:
#                 Distance = _stream.Distance

#             # Operate on the output space Y, Sigma
#             class output:
#                 A_Optimal=_stream.A_Optimal
#                 D_Optimal=_stream.D_Optimal
#                 E_Optimal=_stream.E_Optimal

#             # TODO: add hybrid class (considers X, Y, Sigma)

#             # Adaptive versions of the above
#             class adaptive:
#                 class input:
#                     Distance=_stream.Adaptive_Distance
#                 class output:
#                     A_Optimal=_stream.Adaptive_A_Optimal
#                     D_Optimal=_stream.Adaptive_D_Optimal
#                     E_Optimal=_stream.Adaptive_E_Optimal

__all__ = ["Emulator", "Simulator", "stream"]



import logging
import os
import sys
from pathlib import Path


def configure_logging(log_to_file=False, level: str = "INFO"):
    """
    Configure the logging system.

    Parameters
    ----------
    log_to_file: bool or string, optional
        If True, logs will be written to a file.
        If a string, logs will be written to the specified file.
    verbose: str, optional
        The verbosity level. Can be "error", "warning",
          "info", or "debug". Defaults to "info".
    """
    logger = logging.getLogger("autoemulate")
    logger.handlers = []  # Clear existing handlers

    logger.setLevel(logging.DEBUG)

    verbose_lower = level.lower()
    match verbose_lower:
        case "error":
            console_log_level = logging.ERROR
        case "warning":
            console_log_level = logging.WARNING
        case "info":
            console_log_level = logging.INFO
        case "debug":
            console_log_level = logging.DEBUG
        case _:
            msg = 'verbose must be "error", "warning", "info", or "debug"'
            raise ValueError(msg)

    # Create console handler with a higher log level
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(console_log_level)

    # Create formatter and add it to the handler
    # formatter = logging.Formatter("%(name)s - %(message)s")
    formatter = logging.Formatter("%(levelname)-8s%(asctime)s - %(name)s - %(message)s")
    ch.setFormatter(formatter)

    # Add the handler to the logger
    logger.addHandler(ch)

    # Optionally log to a file
    if log_to_file:
        if isinstance(log_to_file, bool):
            log_file_path = Path.cwd() / "autoemulate.log"
        else:
            log_file_path = Path(log_to_file)
        # Create the directory if it doesn't exist
        log_file_path.parent.mkdir(parents=True, exist_ok=True)

        log_file_dir = os.path.dirname(log_file_path)
        if log_file_dir and not os.path.exists(log_file_dir):
            os.makedirs(log_file_dir)

        try:
            fh = logging.FileHandler(log_file_path)
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(formatter)
            logger.addHandler(fh)
        except Exception:
            logger.exception("Failed to create log file at %s", log_file_path)

    # Capture (model) warnings and redirect them to the logging system
    logging.captureWarnings(True)

    warnings_logger = logging.getLogger("py.warnings")
    for handler in logger.handlers:
        warnings_logger.addHandler(handler)
    warnings_logger.setLevel(logger.getEffectiveLevel())

    return logger


def get_configured_logger(
    log_level, progress_bar_attr="progress_bar"
) -> tuple[logging.Logger, bool]:
    """
    Configure logger and progress bar flag consistently.

    Parameters
    ----------
    log_level: str
        The logging level to set. Can be "progress_bar", "debug", "info",
        "warning", "error", or "critical".
    progress_bar_attr: str
        The attribute to check for progress bar. If log_level is set to this value,
        the logger will be set to "error" level and progress_bar will be True. Defaults
        to "progress_bar".

    Returns
    -------
    tuple[logging.Logger, bool]
        The configured logger and the progress bar flag.
    """
    valid_log_levels = [
        "progress_bar",
        "debug",
        "info",
        "warning",
        "error",
        "critical",
    ]
    log_level = log_level.lower()
    if log_level not in valid_log_levels:
        raise ValueError(
            f"Invalid log level: {log_level}. Must be one of: {valid_log_levels}"
        )
    if log_level == progress_bar_attr:
        log_level = "error"
        progress_bar = True
    else:
        progress_bar = False
    logger = configure_logging(level=log_level)
    return logger, progress_bar



import random

import numpy as np
from torchrbf import RBFInterpolator

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import DeviceLike, OutputLike, TensorLike
from autoemulate.emulators.base import PyTorchBackend
from autoemulate.transforms.standardize import StandardizeTransform


class RadialBasisFunctions(PyTorchBackend):
    """
    Radial basis function Emulator.

    Wraps the Radial Basis Function Interpolation in PyTorch.
    """

    supports_grad = False

    def __init__(  # noqa: PLR0913
        self,
        x: TensorLike,  # noqa: ARG002
        y: TensorLike,  # noqa: ARG002
        standardize_x: bool = False,
        standardize_y: bool = False,
        smoothing: float = 0.0,
        kernel: str = "thin_plate_spline",
        epsilon: float = 1.0,
        degree: int = 1,
        device: DeviceLike | None = None,
    ):
        """Initialize a RadialBasisFunctions emulator.

        Parameters
        ----------
        x: TensorLike
            Input features.
        y: TensorLike
            Target values.
        standardize_x: bool
            Whether to standardize input features. Defaults to False.
        standardize_y: bool
            Whether to standardize target values. Defaults to False.
        smoothing: float
            Smoothing parameter for the RBF interpolator. Defaults to 0.0.
        kernel: str
            Kernel type for the RBF interpolator.
            Kernel type for the RBF interpolator. Options are:
            "linear", "multiquadric", "thin_plate_spline", "cubic", "quintic",
            "gaussian".
            Defaults to "thin_plate_spline".
        epsilon: float
            Epsilon parameter for the RBF interpolator. Defaults to 1.0.
        degree: int
            Degree of the polynomial to be added to the RBF interpolator. Defaults to 1.
        device: DeviceLike | None
            Device to run the model on. If None, uses the default device. Defaults to
            None.
        """
        super().__init__()
        TorchDeviceMixin.__init__(self, device=device)

        self.x_transform = StandardizeTransform() if standardize_x else None
        self.y_transform = StandardizeTransform() if standardize_y else None
        self.smoothing = smoothing
        self.kernel = kernel
        self.epsilon = epsilon
        self.degree = degree
        self.device = device

    def _fit(self, x: TensorLike, y: TensorLike):
        self.model = RBFInterpolator(
            x,
            y,
            smoothing=self.smoothing,
            kernel=self.kernel,
            epsilon=self.epsilon,
            degree=self.degree,
            device=self.device,  # type: ignore PGH003
        )

    def forward(self, x: TensorLike) -> TensorLike:
        """Forward pass for the radial basis function emulator."""
        return self.model(x)

    def _predict(self, x: TensorLike, with_grad: bool) -> OutputLike:
        if with_grad:
            msg = "Gradient calculation is not supported."
            raise ValueError(msg)
        self.eval()
        return self(x)

    @staticmethod
    def is_multioutput() -> bool:
        """Radial basis functions support multi-output."""
        return True

    @staticmethod
    def get_tune_params():
        """Return a dictionary of hyperparameters to tune."""
        all_params = [
            {
                "kernel": ["linear", "multiquadric"],
                "degree": [np.random.randint(0, 3)],  # Degrees valid for these kernels
                "smoothing": [np.random.uniform(0.0, 1.0)],
            },
            {
                "kernel": ["thin_plate_spline", "cubic"],
                "degree": [np.random.randint(1, 3)],
                "smoothing": [np.random.uniform(0.0, 1.0)],
            },
            {
                "kernel": ["quintic"],
                "degree": [np.random.randint(2, 3)],
                "smoothing": [np.random.uniform(0.0, 1.0)],
            },
            {
                "kernel": ["gaussian"],
                "degree": [np.random.randint(-1, 3)],
                "smoothing": [np.random.uniform(0.0, 1.0)],
            },
        ]
        # Randomly select one of the parameter sets
        return random.choice(all_params)



import numpy as np
from lightgbm import LGBMRegressor
from scipy.sparse import spmatrix

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import DeviceLike, TensorLike
from autoemulate.emulators.base import DeterministicEmulator
from autoemulate.transforms.standardize import StandardizeTransform


class LightGBM(DeterministicEmulator):
    """LightGBM Emulator.

    Wraps LightGBM regression from LightGBM.
    See https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMRegressor.html
    for more details.
    """

    def __init__(  # noqa: PLR0913 allow too many arguments since all currently required
        self,
        x: TensorLike | None = None,
        y: TensorLike | None = None,
        standardize_x: bool = False,
        standardize_y: bool = False,
        boosting_type: str = "gbdt",
        num_leaves: int = 31,
        max_depth: int = -1,
        learning_rate: float = 0.1,
        n_estimators: int = 100,
        subsample_for_bin: int = 200000,
        objective: str | None = None,
        class_weight: dict | str | None = None,
        min_split_gain: float = 0.0,
        min_child_weight: float = 0.001,
        min_child_samples: int = 20,
        subsample: float = 1.0,
        colsample_bytree: float = 1.0,
        reg_alpha: float = 0.0,
        reg_lambda: float = 0.0,
        random_seed: int | None = None,
        n_jobs: int | None = 1,
        importance_type: str = "split",
        verbose: int = -1,
        device: DeviceLike = "cpu",
    ):
        """Initialize a LightGBM object.

        Parameters
        ----------
        x: TensorLike | None
            Input features. If None, the model will be fitted later. Defaults to None.
        y: TensorLike | None
            Target values. If None, the model will be fitted later. Defaults to None.
        standardize_x: bool
            Whether to standardize input features. Defaults to None.
        standardize_y: bool
            Whether to standardize target values. Defaults to None.
        boosting_type: str
            Type of boosting to use. Options are "gbdt", "dart", "goss", "rf". Defaults
            to "gbdt".
        num_leaves: int
            Maximum number of leaves in one tree. Defaults to 31.
        max_depth: int
            Maximum depth of the tree. -1 means no limit. Defaults to -1.
        learning_rate: float
            Learning rate shrinks the contribution of each new tree. Defaults to 0.1.
        n_estimators: int
            The number of boosting stages to be run. Defaults to 100.
        subsample_for_bin: int
            Number of samples for constructing bins. Defaults to 200000.
        objective: str | None
            Objective function to be optimized. If None, defaults to "regression".
            Defaults to None.
        class_weight: dict | str | None
            Class weights for multi-class classification. If None, all classes are
            assumed to have equal weight. Defaults to None.
        min_split_gain: float
            Minimum loss reduction required to make a further partition on a leaf node.
            Defaults to 0.0.
        min_child_weight: float
            Minimum sum of instance weight (hessian) needed in a child.
            Defaults to 0.001.
        min_child_samples: int
            Minimum number of data points in a child. Defaults to 20.
        subsample: float
            Fraction of samples to be used for fitting the individual base learners.
            Defaults to 1.0.
        colsample_bytree: float
            Fraction of features to be used for fitting the individual base learners.
            Defaults to 1.0.
        reg_alpha: float
            L1 regularization term on weights. Defaults to 0.0.
        reg_lambda: float
            L2 regularization term on weights. Defaults to 0.0.
        random_seed: int | None
            Random seed for reproducibility. If None, no seed is set.
        n_jobs: int | None
            Number of parallel threads used to run LightGBM. If None, uses all available
            cores. Defaults to 1.
        importance_type: str
            Type of feature importance to be calculated. Options are "split", "gain",
            "cover", "total_gain", "total_cover". Defaults to "split".
        verbose: int
            Verbosity of the output. -1 means no output, 0 means warnings only,
            1 means info, 2 means debug.
        device: DeviceLike
            Device to run the model on (e.g., "cpu", "cuda", "mps"). Defaults to "cpu".
        """
        _, _ = x, y  # ignore unused arguments
        TorchDeviceMixin.__init__(self, device=device)
        self.x_transform = StandardizeTransform() if standardize_x else None
        self.y_transform = StandardizeTransform() if standardize_y else None
        self.boosting_type = boosting_type
        self.num_leaves = num_leaves
        self.max_depth = max_depth
        self.learning_rate = learning_rate
        self.n_estimators = n_estimators
        self.subsample_for_bin = subsample_for_bin
        self.objective = objective
        self.class_weight = class_weight
        self.min_split_gain = min_split_gain
        self.min_child_weight = min_child_weight
        self.min_child_samples = min_child_samples
        self.subsample = subsample
        self.colsample_bytree = colsample_bytree
        self.reg_alpha = reg_alpha
        self.reg_lambda = reg_lambda
        self.random_seed = random_seed
        self.n_jobs = n_jobs
        self.importance_type = importance_type
        self.verbose = verbose
        self.supports_grad = False
        self.model_ = LGBMRegressor(
            boosting_type=self.boosting_type,
            num_leaves=self.num_leaves,
            max_depth=self.max_depth,
            learning_rate=self.learning_rate,
            n_estimators=self.n_estimators,
            subsample_for_bin=self.subsample_for_bin,
            objective=self.objective,
            class_weight=self.class_weight,
            min_split_gain=self.min_split_gain,
            min_child_weight=self.min_child_weight,
            min_child_samples=self.min_child_samples,
            subsample=self.subsample,
            colsample_bytree=self.colsample_bytree,
            reg_alpha=self.reg_alpha,
            reg_lambda=self.reg_lambda,
            random_state=self.random_seed,
            n_jobs=self.n_jobs,
            importance_type=self.importance_type,
            verbose=self.verbose,
        )

    @staticmethod
    def is_multioutput() -> bool:
        """LightGBM does not support multi-output."""
        return False

    def _fit(self, x: TensorLike, y: TensorLike):
        x_np, y_np = self._convert_to_numpy(x, y)
        self.n_features_in_ = x_np.shape[1]
        self.model_.fit(x_np, y_np)

    def _predict(self, x: TensorLike, with_grad: bool) -> TensorLike:
        if with_grad:
            msg = "Gradient calculation is not supported."
            raise ValueError(msg)
        y_pred = self.model_.predict(x)
        assert not isinstance(y_pred, spmatrix | list)
        _, y = self._convert_to_tensors(x, y_pred)
        return y

    @staticmethod
    def get_tune_params():
        """Return a dictionary of hyperparameters to tune."""
        # Note: 10 ** np.random.uniform(-3, 0)
        # is equivalent to scipy.stats.loguniform(0.001, 0.1)
        return {
            "num_leaves": [np.random.randint(10, 100)],
            "max_depth": [np.random.randint(-1, 12)],
            "learning_rate": [10 ** np.random.uniform(-3, -1)],
            "n_estimators": [np.random.randint(50, 1000)],
            "reg_alpha": [10 ** np.random.uniform(-3, 0)],
            "reg_lambda": [10 ** np.random.uniform(-3, 0)],
        }



import inspect
import logging
from functools import partial

import torch
import torchmetrics
from sklearn.model_selection import BaseCrossValidator
from torch.utils.data import Dataset, Subset

from autoemulate.core.device import (
    get_torch_device,
    move_tensors_to_device,
)
from autoemulate.core.types import (
    DeviceLike,
    DistributionLike,
    InputLike,
    ModelParams,
    OutputLike,
    TensorLike,
)
from autoemulate.data.utils import ConversionMixin, set_random_seed
from autoemulate.emulators.base import Emulator
from autoemulate.emulators.transformed.base import TransformedEmulator
from autoemulate.transforms.base import AutoEmulateTransform

logger = logging.getLogger("autoemulate")


def r2_metric() -> type[torchmetrics.Metric]:
    """Return a torchmetrics.R2Score metric."""
    return torchmetrics.R2Score


def rmse_metric() -> partial[torchmetrics.Metric]:
    """Return a torchmetrics.MeanSquaredError metric with squared=False."""
    return partial(torchmetrics.MeanSquaredError, squared=False)


def _update(
    y_true: InputLike,
    y_pred: OutputLike,
    metric: torchmetrics.Metric,
):
    # handle types
    if isinstance(y_pred, TensorLike):
        metric.to(y_pred.device)
        metric.update(y_pred, y_true)
    elif isinstance(y_pred, DistributionLike):
        metric.to(y_pred.mean.device)
        metric.update(y_pred.mean, y_true)
    else:
        raise ValueError(f"Metric not implmented for {type(y_pred)}")


def evaluate(
    y_pred: OutputLike,
    y_true: InputLike,
    metric: (
        type[torchmetrics.Metric] | partial[torchmetrics.Metric]
    ) = torchmetrics.R2Score,
) -> float:
    """
    Evaluate Emulator prediction performance using a `torchmetrics.Metric`.

    Parameters
    ----------
    y_true: InputLike
        Ground truth target values.
    y_pred: OutputLike
        Predicted target values, as returned by an Emulator.
    metric: type[Metric]
        A torchmetrics metric to compute the score.

    Returns
    -------
    float
    """
    _metric = metric()
    _update(y_true, y_pred, _metric)
    return _metric.compute().item()


def cross_validate(  # noqa: PLR0913
    cv: BaseCrossValidator,
    dataset: Dataset,
    model: type[Emulator],
    model_params: ModelParams,
    x_transforms: list[AutoEmulateTransform] | None = None,
    y_transforms: list[AutoEmulateTransform] | None = None,
    device: DeviceLike = "cpu",
    random_seed: int | None = None,
):
    """
    Cross validate model performance using the given `cv` strategy.

    Parameters
    ----------
    cv: BaseCrossValidator
        Provides split method that returns train/val Dataset indices using the
        specified cross-validation strategy (e.g., KFold, LeaveOneOut).
    dataset: Dataset
        The data to use for model training and validation.
    model: Emulator
        An instance of an Emulator subclass.
    model_params: ModelParams
        Model parameters to be used to construct model upon initialization. Passing an
        empty dictionary `{}` will use default parameters.
    device: DeviceLike
        The device to use for model training and evaluation.
    random_seed: int | None
        Optional random seed for reproducibility.

    Returns
    -------
    dict[str, list[float]]
       Contains r2 and rmse scores computed for each cross validation fold.
    """
    x_transforms = x_transforms or []
    y_transforms = y_transforms or []
    cv_results = {"r2": [], "rmse": []}
    device = get_torch_device(device)

    logger.debug("Cross-validation parameters: %s", cv)
    for i, (train_idx, val_idx) in enumerate(cv.split(dataset)):  # type: ignore TODO: identify type handling here
        logger.debug(
            "Cross-validation split %d: %d train samples, %d validation samples",
            i,
            len(train_idx),
            len(val_idx),
        )

        # create train/val data subsets
        # convert idx to list to satisfy type checker
        train_subset = Subset(dataset, train_idx.tolist())
        val_subset = Subset(dataset, val_idx.tolist())

        # Handle random seed for reproducibility
        if random_seed is not None:
            set_random_seed(seed=random_seed)
        model_init_params = inspect.signature(model).parameters
        model_kwargs = dict(model_params)
        if "random_seed" in model_init_params:
            model_kwargs["random_seed"] = random_seed

        # Convert dataloader to tensors to pass to model
        x, y = ConversionMixin._convert_to_tensors(train_subset)
        x_val, y_val = ConversionMixin._convert_to_tensors(val_subset)

        transformed_emulator = TransformedEmulator(
            x,
            y,
            model=model,
            x_transforms=x_transforms,
            y_transforms=y_transforms,
            device=device,
            **model_kwargs,
        )
        transformed_emulator.fit(x, y)

        # compute and save results
        y_pred = transformed_emulator.predict(x_val)
        r2 = evaluate(y_pred, y_val, r2_metric())
        rmse = evaluate(y_pred, y_val, rmse_metric())
        cv_results["r2"].append(r2)
        cv_results["rmse"].append(rmse)
    return cv_results


def bootstrap(
    model: Emulator,
    x: TensorLike,
    y: TensorLike,
    n_bootstraps: int = 100,
    device: str | torch.device = "cpu",
) -> tuple[tuple[float, float], tuple[float, float]]:
    """
    Get bootstrap estimates of R2 and RMSE.

    Parameters
    ----------
    model: Emulator
        An instance of an Emulator subclass.
    x: TensorLike
        Input features for the model.
    y: TensorLike
        Target values corresponding to the input features.
    n_bootstraps: int
        Number of bootstrap samples to generate. Defaults to 100.
    device: str | torch.device
        The device to use for computations. Default is "cpu".

    Returns
    -------
    tuple[tuple[float, float], tuple[float, float]]
        ((r2_mean, r2_std), (rmse_mean, rmse_std))
    """
    device = get_torch_device(device)
    x, y = move_tensors_to_device(x, y, device=device)

    r2_scores = torch.empty(n_bootstraps, device=device)
    rmse_scores = torch.empty(n_bootstraps, device=device)
    for i in range(n_bootstraps):
        # Bootstrap sample indices with replacement
        idxs = torch.randint(0, x.shape[0], size=(x.shape[0],), device=device)

        # Get bootstrap sample
        x_bootstrap = x[idxs]
        y_bootstrap = y[idxs]

        # Make predictions
        y_pred = model.predict(x_bootstrap)

        # Compute metrics for this bootstrap sample
        r2_scores[i] = evaluate(y_pred, y_bootstrap, r2_metric())
        mse_score = evaluate(y_pred, y_bootstrap, rmse_metric())
        rmse_scores[i] = mse_score**0.5

    # Return mean and std
    return (
        (r2_scores.mean().item(), r2_scores.std().item()),
        (rmse_scores.mean().item(), rmse_scores.std().item()),
    )



import warnings

import torch
from linear_operator.utils.warnings import NumericalWarning


def make_positive_definite(
    cov, min_jitter: float = 1e-6, max_tries: int = 3, clamp_eigvals: bool = False
):
    """
    Ensure a covariance matrix is positive definite.

    Ensure a covariance matrix is positive definite by:
        1. adding increasing amounts of jitter to the diagonal and symmetrizing
           the matrix
        2. if this fails, clamping eigenvalues to a minimum value
        3. if this still fails, clamping eigenvalues to both minimum and maximum
           values based on the median eigenvalue

    See related function in linear_operator: `psd_safe_cholesky`:
    https://github.com/cornellius-gp/linear_operator/blob/dca438e47dd8a380d0f4e6b30c406e187062c8bd/linear_operator/utils/cholesky.py#L12

    Parameters
    ----------
    cov: torch.Tensor
        The covariance matrix to be made positive definite. Must be a 2D or 3D tensor.
    min_jitter: float
        Starting value for jitter to add to the diagonal. Jitter is increased by
        powers of 10 for each attempt. Defaults to 1e-6.
    max_tries: int
        Number of attempts to add jitter before moving to eigenvalue clamping
        (if enabled) or raising an error. Defaults to 3.
    clamp_eigvals: bool
        Whether to attempt eigenvalue clamping if adding jitter fails. If False,
        will raise an error after max_tries jitter attempts. Defaults to False.

    Returns
    -------
    torch.Tensor
        A positive definite covariance matrix of the same shape as input.

    Raises
    ------
    ValueError
        If cov is not a 2D or 3D tensor.
    RuntimeError
        If the matrix cannot be made positive definite after all attempts.
    """
    # If already positive definite, return as is
    try:
        torch.linalg.cholesky(cov)
        return cov
    except RuntimeError:
        pass

    # If that fails, try adding increasing amounts of jitter
    jitter = min_jitter
    for i in range(max_tries):
        # Add jitter to diagonal and ensure symmetry
        eye = torch.eye(cov.shape[-1], device=cov.device, dtype=cov.dtype)
        if len(cov.shape) == 3:
            eye = eye.unsqueeze(0)
        if len(cov.shape) > 3:
            msg = f"cov must be a 2D or 3D tensor, got shape {cov.shape}"
            raise ValueError(msg) from None
        cov = cov + jitter * eye
        cov = (cov + cov.transpose(-1, -2)) / 2

        try:
            torch.linalg.cholesky(cov)
            warnings.warn(
                f"cov not p.d. - added {jitter:.1e} to the diagonal and symmetrized",
                NumericalWarning,
                stacklevel=2,
            )
            return cov
        except RuntimeError as e:
            if i == max_tries - 1 and not clamp_eigvals:
                msg = (
                    f"Matrix could not be made positive definite after "
                    f"{max_tries} attempts with jitter up to {jitter:.1e}:\n{cov}"
                )
                raise RuntimeError(msg) from e
            # Increase jitter for next attempt
            jitter *= 10

    # If adding jitter fails, optionally clamp eigvals to find closest positive definite
    # matrix using heuristic for upper bound of eigenvalues if clamping lower bound is
    # not sufficient
    MIN_EIGVAL = 1e-3
    if clamp_eigvals:
        # Attempt to first only clamp minimum eigenvals
        cov = (cov + cov.transpose(-1, -2)) / 2
        eigvals, eigvecs = torch.linalg.eigh(cov)
        eigvals = torch.clamp(eigvals, min=MIN_EIGVAL)
        cov = eigvecs @ torch.diag_embed(eigvals) @ eigvecs.transpose(-1, -2)
        cov = (cov + cov.transpose(-1, -2)) / 2
        try:
            torch.linalg.cholesky(cov)
            warnings.warn(
                f"cov not p.d. - clamped min eigval to {MIN_EIGVAL:.1e} and "
                "symmetrized",
                NumericalWarning,
                stacklevel=2,
            )
            return cov
        except RuntimeError:
            # Clamp both minimum and maximum eigvals
            cov = (cov + cov.transpose(-1, -2)) / 2
            eigvals, eigvecs = torch.linalg.eigh(cov)

            # Ensure condition number around 10**6
            min_eigval = max(torch.min(eigvals).item(), MIN_EIGVAL)
            max_eigval = min(torch.max(eigvals).item(), 1e6 * min_eigval)

            # Clamp eigvals
            eigvals = torch.clamp(eigvals, min=min_eigval, max=max_eigval)
            cov = eigvecs @ torch.diag_embed(eigvals) @ eigvecs.transpose(-1, -2)
            cov = (cov + cov.transpose(-1, -2)) / 2
            try:
                torch.linalg.cholesky(cov)
                warnings.warn(
                    f"cov not p.d. - clamped min eigval to {min_eigval:.1e} and max "
                    f"eigval to {max_eigval:.1e}, then symmetrized",
                    NumericalWarning,
                    stacklevel=2,
                )
                return cov
            except RuntimeError as e:
                msg = (
                    f"Matrix could not be made positive definite after clamping "
                    f"eigenvalues to min={min_eigval:.1e} and max={max_eigval:.1e}:\n"
                    f"{cov}"
                )
                raise RuntimeError(msg) from e

    raise RuntimeError(f"Matrix could not be made positive definite:\n{cov}")



import torch
from gpytorch.kernels import (
    ConstantKernel,
    Kernel,
    LinearKernel,
    MaternKernel,
    RBFKernel,
    RQKernel,
)


def rbf(n_features: int | None, n_outputs: torch.Size | None) -> RBFKernel:
    """
    Radial Basis Function (RBF) kernel.

    Parameters
    ----------
    n_features: int | None
        Number of input features. If None, the kernel is not initialized with a
        lengthscale.
    n_outputs: torch.Size | None
        Batch shape of the kernel. If None, the kernel is not initialized with a
        batch shape.

    Returns
    -------
    RBFKernel
        The initialized RBF kernel.
    """
    k = RBFKernel(
        ard_num_dims=n_features,
        batch_shape=n_outputs,
    )
    return (
        k.initialize(lengthscale=torch.ones(n_features) * 1.5)
        if n_features is not None
        else k
    )


def matern_5_2_kernel(
    n_features: int | None, n_outputs: torch.Size | None
) -> MaternKernel:
    """
    Matern 5/2 kernel.

    Parameters
    ----------
    n_features: int | None
        Number of input features. If None, the kernel is not initialized with a
        lengthscale.
    n_outputs: torch.Size | None
        Batch shape of the kernel. If None, the kernel is not initialized with a
        batch shape.

    Returns
    -------
    MaternKernel
        The initialized Matern 5/2 kernel.
    """
    k = MaternKernel(
        nu=2.5,
        ard_num_dims=n_features,
        batch_shape=n_outputs,
    )
    return (
        k.initialize(lengthscale=torch.ones(n_features) * 1.5)
        if n_features is not None
        else k
    )


def matern_3_2_kernel(
    n_features: int | None, n_outputs: torch.Size | None
) -> MaternKernel:
    """
    Matern 3/2 kernel.

    Parameters
    ----------
    n_features: int | None
        Number of input features. If None, the kernel is not initialized with a
        lengthscale.
    n_outputs: torch.Size | None
        Batch shape of the kernel. If None, the kernel is not initialized with a
        batch shape.

    Returns
    -------
    MaternKernel
        The initialized Matern 3/2 kernel.

    """
    k = MaternKernel(
        nu=1.5,
        ard_num_dims=n_features,
        batch_shape=n_outputs,
    )
    return (
        k.initialize(lengthscale=torch.ones(n_features) * 1.5)
        if n_features is not None
        else k
    )


def rq_kernel(n_features: int | None, n_outputs: torch.Size | None) -> RQKernel:
    """
    Rational Quadratic (RQ) kernel.

    Parameters
    ----------
    n_features: int | None
        Number of input features. If None, the kernel is not initialized with a
        lengthscale.
    n_outputs: torch.Size | None
        Batch shape of the kernel. If None, the kernel is not initialized with a
        batch shape.

    Returns
    -------
    RQKernel
        The initialized RQ kernel.
    """
    k = RQKernel(
        ard_num_dims=n_features,
        batch_shape=n_outputs,
    )
    return (
        k.initialize(lengthscale=torch.ones(n_features) * 1.5)
        if n_features is not None
        else k
    )


def rbf_plus_constant(n_features: int | None, n_outputs: torch.Size | None) -> Kernel:
    """
    Radial Basis Function (RBF) kernel plus a constant kernel.

    Parameters
    ----------
    n_features: int | None
        Number of input features. If None, the kernel is not initialized with a
        lengthscale.
    n_outputs: torch.Size | None
        Batch shape of the kernel. If None, the kernel is not initialized with a
        batch shape.

    Returns
    -------
    Kernel
        The initialized RBF kernel plus a constant kernel.
    """
    rbf_kernel = RBFKernel(
        ard_num_dims=n_features,
        batch_shape=n_outputs,
    )
    if n_features is not None:
        rbf_kernel.initialize(lengthscale=torch.ones(n_features) * 1.5)
    return rbf_kernel + ConstantKernel()


# combinations
def rbf_plus_linear(n_features: int | None, n_outputs: torch.Size | None) -> Kernel:
    """
    Radial Basis Function (RBF) kernel plus a linear kernel.

    Parameters
    ----------
    n_features: int | None
        Number of input features. If None, the kernel is not initialized with a
        lengthscale.
    n_outputs: torch.Size | None
        Batch shape of the kernel. If None, the kernel is not initialized with a
        batch shape.

    Returns
    -------
    Kernel
        The initialized RBF kernel plus a linear kernel.
    """
    rbf_kernel = RBFKernel(
        ard_num_dims=n_features,
        batch_shape=n_outputs,
    )
    if n_features is not None:
        rbf_kernel.initialize(lengthscale=torch.ones(n_features) * 1.5)
    return rbf_kernel + LinearKernel(
        ard_num_dims=n_features,
        batch_shape=n_outputs,
    )


def matern_5_2_plus_rq(n_features: int | None, n_outputs: torch.Size | None) -> Kernel:
    """
    Matern 5/2 kernel plus a Rational Quadratic (RQ) kernel.

    Parameters
    ----------
    n_features: int | None
        Number of input features. If None, the kernel is not initialized with a
        lengthscale.
    n_outputs: torch.Size | None
        Batch shape of the kernel. If None, the kernel is not initialized with a
        batch shape.

    Returns
    -------
    Kernel
        The initialized Matern 5/2 kernel plus a Rational Quadratic kernel.
    """
    matern_kernel = MaternKernel(
        nu=2.5,
        ard_num_dims=n_features,
        batch_shape=n_outputs,
    )
    rq_kernel = RQKernel(
        ard_num_dims=n_features,
        batch_shape=n_outputs,
    )
    # Initialize lengthscales for both kernels if n_features is provided
    if n_features is not None:
        matern_kernel.initialize(lengthscale=torch.ones(n_features) * 1.5)
        rq_kernel.initialize(lengthscale=torch.ones(n_features) * 1.5)
    return matern_kernel + rq_kernel


def rbf_times_linear(n_features: int | None, n_outputs: torch.Size | None) -> Kernel:
    """Radial Basis Function (RBF) kernel multiplied by a linear kernel.

    Parameters
    ----------
    n_features: int | None
        Number of input features. If None, the kernel is not initialized with a
        lengthscale.
    n_outputs: torch.Size | None
        Batch shape of the kernel. If None, the kernel is not initialized with a
        batch shape.

    Returns
    -------
    Kernel
        The initialized RBF kernel multiplied by a linear kernel.
    """
    rbf_kernel = RBFKernel(
        ard_num_dims=n_features,
        batch_shape=n_outputs,
    )
    if n_features is not None:
        rbf_kernel.initialize(lengthscale=torch.ones(n_features) * 1.5)
    return rbf_kernel * LinearKernel(
        ard_num_dims=n_features,
        batch_shape=n_outputs,
    )



from abc import ABC, abstractmethod

import mogp_emulator
import torch

from autoemulate.core.types import TensorLike


class ExperimentalDesign(ABC):
    """
    Abstract class for experimental design.

    This class is used to sample points from the parameter space to be used in the
    training of an emulator.
    """

    @abstractmethod
    def __init__(self, bounds_list: list[tuple[float, float]]):
        """
        Initialize a Sampler object.

        Parameters
        ----------
        bounds_list: list
            List tuples with two numeric values. Each tuple corresponds to the lower and
            upper bounds of a parameter.
        """

    @abstractmethod
    def sample(self, n: int) -> TensorLike:
        """
        Sample n points from the sample space.

        Parameters
        ----------
        n: int
            The number of points to sample.

        Returns
        -------
        TensorLike
            A tensor of shape (n, dim) containing the sampled points.
        """

    @abstractmethod
    def get_n_parameters(self) -> int:
        """
        Return the number of parameters in the sample space.

        Returns
        -------
        int
            The number of parameters in the sample space.
        """


class LatinHypercube(ExperimentalDesign):
    """LatinHypercube experimental design class."""

    def __init__(self, bounds_list: list[tuple[float, float]]):
        """Initialize a LatinHypercube object."""
        self.bounds_list = bounds_list

        # Partition indices and bounds into differing and constant parameters
        self.diff_idxs = []
        self.same_idxs = []
        bounds_list_diff = []
        for idx, bounds in enumerate(bounds_list):
            if bounds[0] != bounds[1]:
                self.diff_idxs.append(idx)
                bounds_list_diff.append(bounds)
            else:
                self.same_idxs.append(idx)

        self.sampler = mogp_emulator.LatinHypercubeDesign(bounds_list_diff)

    def sample(self, n: int) -> TensorLike:
        """
        Sample n points from the parameter space.

        Parameters
        ----------
        n: int
            The number of points to sample.

        Returns
        -------
        TensorLike
            A tensor of shape (n, dim) containing the sampled points.
        """
        # Sample only from parameters that differ
        sample_array_diff = self.sampler.sample(n)

        # Create full sample array filling in constant parameters
        sample_array_full = torch.zeros((n, len(self.bounds_list)), dtype=torch.float32)
        sample_array_full[:, self.diff_idxs] = torch.tensor(
            sample_array_diff, dtype=torch.float32
        )
        for idx in self.same_idxs:
            sample_array_full[:, idx] = self.bounds_list[idx][0]

        return sample_array_full

    def get_n_parameters(self) -> int:
        """
        Return the number of parameters in the sample space.

        Returns
        -------
        int
            The number of parameters in the sample space.
        """
        return len(self.bounds_list)



import inspect
from abc import ABC, abstractmethod
from collections.abc import Callable

import torch
from linear_operator.operators import DiagLinearOperator
from torch.distributions import Transform
from typing_extensions import Self

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import (
    DistributionLike,
    GaussianLike,
    GaussianProcessLike,
    TensorLike,
)
from autoemulate.data.utils import ConversionMixin, ValidationMixin
from autoemulate.transforms.utils import make_positive_definite


class AutoEmulateTransform(
    Transform, ABC, ValidationMixin, ConversionMixin, TorchDeviceMixin
):
    """
    Base class for transforms in the AutoEmulate framework.

    This class subclasses the `torch.distributions.Transform` class and provides
    additional functionality for fitting transforms to data and transforming
    Gaussian distributions between the codomain and domain of the transform.

    """

    _is_fitted: bool = False

    # TODO: consider if the override also needs to consider DistributionLike case
    def __call__(self, x: TensorLike) -> TensorLike:
        """Apply the transform to input tensor `x`."""
        output = super().__call__(x)
        assert isinstance(output, TensorLike)
        return output

    @abstractmethod
    def fit(self, x: TensorLike):
        """Fit the transform to the input tensor `x`."""
        ...

    def _check_is_fitted(self):
        """Check if the transform has been fitted and otherwise raise an error."""
        if not self._is_fitted:
            msg = f"Transform ({self}) has not been fitted yet."
            raise RuntimeError(msg)

    def to_dict(self) -> dict:
        """
        Serialize the transform to a dictionary.

        Returns
        -------
        dict
            A dictionary with transform name as key and initialization parameters
            as value. The format is:
            {
                "transform_name": {param1: value1, param2: value2, ...}
            }
        """
        # Get the transform name (convert class name to lowercase, remove 'Transform')
        class_name = self.__class__.__name__
        if class_name.endswith("Transform"):
            transform_name = class_name[:-9].lower()  # Remove 'Transform' suffix
        else:
            transform_name = class_name.lower()

        # Get initialization parameters from the __init__ signature
        init_signature = inspect.signature(self.__class__.__init__)
        init_params = {}

        for param_name, param in init_signature.parameters.items():
            # Skip 'self' parameter
            if param_name == "self":
                continue

            # Get the value from the instance if it exists
            if hasattr(self, param_name):
                value = getattr(self, param_name)

                # Handle special cases for serialization
                if isinstance(value, torch.Tensor):
                    # Convert tensors to lists for JSON serialization
                    init_params[param_name] = value.tolist()
                elif isinstance(value, torch.device):
                    # Convert device to string
                    init_params[param_name] = str(value)
                else:
                    init_params[param_name] = value
            elif param.default is not inspect.Parameter.empty:
                # Use default value if attribute doesn't exist
                init_params[param_name] = param.default

        return {transform_name: init_params}

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        """
        Deserialize a transform from a dictionary.

        Parameters
        ----------
        data: dict
            A dictionary with transform name as key and initialization parameters
            as value, as created by `to_dict()`.

        Returns
        -------
        AutoEmulateTransform
            An instance of the transform class with the specified parameters.

        Raises
        ------
        ValueError
            If the dictionary format is invalid or the transform cannot be found.

        """
        if len(data) != 1:
            msg = "Dictionary must contain exactly one transform"
            raise ValueError(msg)

        transform_name, init_params = next(iter(data.items()))

        # Get transform class from registry
        transform_class = cls._get_transform_class(transform_name)
        if transform_class is None:
            available = ", ".join(cls._get_available_transforms())
            msg = f"Unknown transform: {transform_name}. Available: {available}"
            raise ValueError(msg)

        # Handle special cases for deserialization
        processed_params = {}
        for param_name, value in init_params.items():
            if param_name == "device" and isinstance(value, str):
                # Convert device string back to torch.device
                if value != "None":
                    processed_params[param_name] = torch.device(value)
                else:
                    processed_params[param_name] = None
            else:
                processed_params[param_name] = value

        return transform_class(**processed_params)

    @classmethod
    def _get_transform_class(cls, transform_name: str):
        """
        Get transform class from the registry.

        Parameters
        ----------
        transform_name: str
            The name of the transform (e.g., 'pca', 'vae', 'standardize')

        Returns
        -------
        type[AutoEmulateTransform] | None
            The transform class if found, None otherwise
        """
        from . import TRANSFORM_REGISTRY  # Lazy import to avoid circular dependency

        return TRANSFORM_REGISTRY.get(transform_name)

    @classmethod
    def _get_available_transforms(cls):
        """
        Get list of available transform names.

        Returns
        -------
        list[str]
            List of available transform names
        """
        from . import TRANSFORM_REGISTRY  # Lazy import to avoid circular dependency

        return list(TRANSFORM_REGISTRY.keys())

    @property
    def _basis_matrix(self) -> TensorLike:
        """
        Constant basis matrix for transforming matrices.

        Subclasses should implement this property (if possible) to return the
        appropriate basis matrix.
        """
        msg = "This method should be implemented in subclasses."
        raise NotImplementedError(msg)

    def _expanded_basis_matrix(self, y: TensorLike) -> TensorLike:
        """
        Return the expanded basis matrix for the transformation.

        Expanded basis matrix for the transformation of the number of samples in
        a given codomain `y` sample tensor.

        Given `n` samples in `y`, this returns a Kronecker product of the identity
        matrix with the basis matrix, effectively expanding the basis matrix to match
        the number of samples in `y`.

        The default implementation assumes that the transform has implemented the
        `_basis_matrix` property for a single sample. However, this can be overridden
        in subclasses to enable alternative approaches to generating the expanded basis
        matrix for a given set of samples.

        Parameters
        ----------
        y: TensorLike
            Input tensor of shape `(n, )` from which to compute the expanded
            basis matrix. The number of samples `n` is inferred from the shape of `y`.

        Returns
        -------
        TensorLike
            Expanded basis matrix.

        Raises
        ------
        RuntimeError
            If the transform has not been fitted yet.
        """
        self._check_is_fitted()
        return torch.kron(torch.eye(y.shape[0], device=self.device), self._basis_matrix)

    def _batch_basis_matrix(self, y: TensorLike) -> TensorLike:
        """
        Return the batch basis matrix for the transformation.

        Batch basis matrix for the transformation of the number of samples in
        a given codomain `y` sample tensor.

        Given `n` samples in `y`, this returns a basis matrix for transforming each
        element of a batch.

        The default implementation assumes that the transform has implemented the
        `_basis_matrix` property for a single sample and is therefore simply repeats
        this along a new batch dim.

        However, the method takes a `y` tensor argument to allow the method to be
        overriden with local approximations such as through the delta method.

        Parameters
        ----------
        y: TensorLike
            Input tensor of shape `(n, )` from which to compute the batch
            basis matrix. The number of samples `n` is inferred from the shape of `y`.

        Returns
        -------
        TensorLike
            Batch basis matrix.

        Raises
        ------
        RuntimeError
            If the transform has not been fitted yet.
        """
        self._check_is_fitted()
        return self._basis_matrix.repeat(y.shape[0], 1, 1)

    def _inverse_sample_gaussian_like(
        self, y: GaussianLike, n_samples: int = 1000, full_covariance: bool = True
    ) -> GaussianLike:
        return _inverse_sample_gaussian_like(
            self.inv, y, n_samples=n_samples, full_covariance=full_covariance
        )

    def _inverse_sample_gaussian_process_like(
        self, y: GaussianLike, n_samples: int = 1000, full_covariance: bool = True
    ) -> GaussianLike:
        return _inverse_sample_gaussian_process_like(
            self.inv, y, n_samples=n_samples, full_covariance=full_covariance
        )

    def _inverse_sample(
        self, y: GaussianLike, n_samples: int = 1000, full_covariance: bool = True
    ) -> GaussianLike:
        """
        Invert a `GaussianLike` distribution by sampling from the codomain.

        Transforms a `GaussianLike` in the codomain to a `GaussianLike` in the domain
        through generating samples from `y` in the codomain and mapping those back
        to the original space `x`.

        The empirical mean and covariance of the samples are computed, and a
        `GaussianLike` object in the domain is returned with these statistics.

        Parameters
        ----------
        y: GaussianLike
            The distribution in the codomain.
        n_samples: int
            Number of samples to generate from the distribution `y`. Defaults to 1000.
        full_covariance: bool
            If True, calculates a full covariance matrix from samples; otherwise,
            calculates only the diagonal of the covariance matrix. This is useful
            for a high-dimensional domain where full covariance might be
            computationally expensive. Defaults to True.

        Returns
        -------
        GaussianLike
            A `GaussianLike` object representing the distribution in the domain with
            mean and covariance derived from the samples.

        Raises
        ------
        RuntimeError
            If covariance matrix cannot be made positive definite.
        """
        if isinstance(y, GaussianProcessLike):
            return self._inverse_sample_gaussian_process_like(
                y, n_samples=n_samples, full_covariance=full_covariance
            )

        return self._inverse_sample_gaussian_like(
            y, n_samples=n_samples, full_covariance=full_covariance
        )

    def _inverse_gaussian(self, y: GaussianLike) -> GaussianLike:
        r"""
        Invert a `GaussianLike` distribution to the original space.

        Transforms a `GaussianLike` in the codomain to a `GaussianLike` in the
        domain by applying the inverse of the transform to the mean and covariance of
        `y` in the codomain.

        The default implementation computes the mean and covariance of `y` and:
        - applies the inverse transformation to the mean
        - transforms the covariance matrix using the expanded basis matrix

        Parameters
        ----------
        y: GaussianLike
            The distribution in the codomain to be transformed back to the domain.

        Returns
        -------
        GaussianLike
            A `GaussianLike` representing the distribution in the domain, with mean and
            covariance derived from the transformed statistics.

        Raises
        ------
        RuntimeError
            If the covariance matrix cannot be made positive definite.

        TypeError
            If the input `y` is not of type `GaussianLike` or `GaussianProcessLike`.

        Notes
        -----
        This method assumes that the transform is either an exactly linear or a
        well-defined approximately linear transformation between the codomain and the
        domain such that the transform can be expressed as follows:

        .. math::
            \mathcal{N}(\mu_y, \Sigma_y)
            \rightarrow
            \mathcal{N}(f^{-1}(\mu_y), \; A \Sigma_y A^\top)

        where :math:`f^{-1}` is the inverse of the transform, :math:`\mu_y` is the mean
        of the distribution in the codomain, :math:`\Sigma_y` is the covariance matrix
        of the distribution in the codomain, and :math:`A` is the linear transformation
        matrix (expanded basis matrix) derived from the transform.
        """
        mean = y.mean
        cov = y.covariance_matrix
        mean_orig = self.inv(mean)
        assert isinstance(mean, TensorLike)
        assert isinstance(cov, TensorLike)
        assert isinstance(mean_orig, TensorLike)

        if isinstance(y, GaussianProcessLike):
            # Get the expanded basis matrix around the mean
            expanded_basis_matrix = self._expanded_basis_matrix(mean)

            # Transform covariance matrix
            cov_orig = expanded_basis_matrix @ cov @ expanded_basis_matrix.T

            # Ensure positive definite
            cov_orig = make_positive_definite(
                cov_orig, min_jitter=1e-6, max_tries=3, clamp_eigvals=True
            )

            return GaussianProcessLike(mean_orig, cov_orig)

        if isinstance(y, GaussianLike):
            if len(y.batch_shape) > 1:
                msg = f"Batch shape ({y.batch_shape}) greater than ndim=1 not supported"
                raise NotImplementedError(msg)
            # Transform covariance matrix
            cov_orig = (
                self._batch_basis_matrix(mean)
                @ cov
                @ self._batch_basis_matrix(mean).transpose(-1, -2)
            )

            # Ensure positive definite
            if cov_orig.ndim > 2:
                # TODO: consider revising whether to only use jitter
                cov_orig = torch.stack(
                    [
                        make_positive_definite(
                            c, min_jitter=1e-6, max_tries=3, clamp_eigvals=True
                        )
                        for c in cov_orig
                    ],
                    0,
                )

            return GaussianLike(mean_orig, cov_orig)

        msg = f"Unsupported type: {type(y)}"
        raise TypeError(msg)


def _inverse_sample_gaussian_like(
    c: Callable,
    y: DistributionLike,
    n_samples: int = 1000,
    full_covariance: bool = True,
) -> GaussianLike:
    """
    Invert a `DistributionLike` to a `GaussianLike` through sampling from `y`.

    Transforms a `DistributionLike` to a `GaussianLike` through sampling from `y`.

    Parameters
    ----------
    c: Callable
        A callable that applies a transformation to the generated samples.
    y: DistributionLike
        The distribution from which to sample.
    n_samples: int
        Number of samples to generate from the distribution `y`. Defaults to 1000.
    full_covariance: bool
        If True, calculates a full covariance matrix from samples; otherwise,
        calculates only the diagonal of the covariance matrix. This is useful
        for a high-dimensional domain where full covariance might be
        computationally expensive. Defaults to True.

    Returns
    -------
    GaussianProcessLike
        A `GaussianProcessLike` object representing the distribution from the empirical
        samples.

    Raises
    ------
    NotImplementedError
        If the batch shape of `y` is greater than 1, as this implementation does
        not support multi-dimensional batch shapes.
    """
    batch_shape = y.batch_shape
    if len(batch_shape) > 1:
        msg = f"Batch shape ({batch_shape}) greater than ndim=1 not supported"
        raise NotImplementedError(msg)

    # Sample from the distribution `y` and apply the transformation `c`
    samples = torch.vmap(c)(torch.stack([y.sample() for _ in range(n_samples)], dim=0))

    # Remove batch dim if batch_shape is empty
    if len(batch_shape) == 0:
        samples = samples.squeeze(1)

    # Ensure the samples are of type TensorLike
    assert isinstance(samples, TensorLike)

    # Compute the mean and covariance of the samples
    mean = samples.mean(dim=0)
    if not full_covariance:
        return GaussianLike(mean, DiagLinearOperator(samples.var(dim=0)))

    cov = (
        # Loop over the batch dimension and compute covariance for each sample
        torch.stack(
            [
                make_positive_definite(
                    s.T.cov(), min_jitter=1e-6, max_tries=3, clamp_eigvals=False
                )
                for s in samples.transpose(0, 1)
            ],
            0,
        )
        if len(batch_shape) > 0
        # If no batch shape, compute covariance for the entire sample set
        else make_positive_definite(
            samples.T.cov(), min_jitter=1e-6, max_tries=3, clamp_eigvals=False
        )
    )
    return GaussianLike(mean, cov)


def _inverse_sample_gaussian_process_like(
    c: Callable,
    y: GaussianProcessLike,
    n_samples: int = 1000,
    full_covariance: bool = True,
) -> GaussianProcessLike:
    """
    Invert a `GaussianProcessLike` distribution by sampling from the codomain.

    Transforms a `GaussianProcessLike` to another `GaussianProcessLike` through
    sampling from `y`.

    Parameters
    ----------
    c: Callable
        A callable that applies a transformation to the generated samples.
    y: GaussianProcessLike
        The Gaussian Process from which to sample.
    n_samples: int
        Number of samples to generate from the distribution `y`. Defaults to 1000.
    full_covariance: bool
        If True, calculates a full covariance matrix from samples; otherwise, calculates
        only the diagonal of the covariance matrix. This is useful for a
        high-dimensional domain where full covariance might be computationally
        expensive. Defaults to True.

    Returns
    -------
    GaussianProcessLike
        A `GaussianProcessLike` object representing the distribution from the empirical
        samples.

    Raises
    ------
    RuntimeError
        If the covariance matrix cannot be made positive definite.
    """
    samples = c(torch.stack([y.sample() for _ in range(n_samples)], dim=0))
    assert isinstance(samples, TensorLike)
    mean = samples.mean(dim=0)
    cov = (
        make_positive_definite(
            samples.reshape(n_samples, -1).T.cov(),
            min_jitter=1e-6,
            max_tries=3,
            clamp_eigvals=True,
        )
        if full_covariance
        else DiagLinearOperator(samples.reshape(n_samples, -1).var(dim=0))
    )
    return GaussianProcessLike(mean, cov)


# TODO (#536): complete implementation
# class AutoEmulateTransformModule(TransformModule):
#     @abstractmethod
#     def fit(self, x): ...
#     def _inverse_gaussian(self, y: GaussianLike) -> GaussianLike:
#         msg = "This method should be implemented in subclasses."
#         raise NotImplementedError(msg)

#     def _inverse_sample(self, y: GaussianLike, n_samples: int = 100) -> GaussianLike:
#         msg = "This method should be implemented in subclasses."
#         raise NotImplementedError(msg)



import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from IPython.display import clear_output, display

from autoemulate.core.types import NumpyLike, TensorLike


class HistoryMatchingDashboard:
    """
    History Matching Dashboard.

    Interactive dashboard for exploring history matching with UI controls that adapt
    based on selected plot type.
    """

    def __init__(
        self,
        samples: NumpyLike | TensorLike,
        impl_scores: NumpyLike | TensorLike,
        param_names: list[str],
        output_names: list[str],
        threshold: float = 3.0,
    ):
        """
        Initialize the dashboard.

        Parameters
        ----------
        samples: NumpyLike or TensorLike
            Parameter samples.
        impl_scores: NumpyLike or TensorLike
            Implausibility scores.
        param_names: list[str]
            Parameter names.
        output_names: list[str]
            Output names.
        threshold: float
            Implausibility threshold.
        """
        # Convert samples to DataFrame
        if isinstance(samples, np.ndarray):
            self.samples_df = pd.DataFrame(samples, columns=param_names)  # pyright: ignore[reportArgumentType]
        elif isinstance(samples, TensorLike):
            self.samples_df = pd.DataFrame(samples.numpy(), columns=param_names)  # pyright: ignore[reportArgumentType]

        # Store other data
        if isinstance(impl_scores, TensorLike):
            self.impl_scores = impl_scores.numpy()
        else:
            self.impl_scores = impl_scores
        self.param_names = param_names
        self.output_names = output_names
        self.threshold = threshold

        # Calculate minimum implausibility for each sample
        if len(self.impl_scores.shape) > 1:
            self.min_impl = np.min(self.impl_scores, axis=1)
            self.max_impl = np.max(self.impl_scores, axis=1)
        else:
            self.min_impl = self.impl_scores
            self.max_impl = self.impl_scores

        # Add implausibility to DataFrame
        self.samples_df["min_implausibility"] = self.min_impl
        self.samples_df["max_implausibility"] = self.max_impl
        self.samples_df["NROY"] = self.max_impl <= threshold

        # Create the UI elements
        self._create_ui()

    def _create_ui(self):
        """Create the user interface widgets with dynamic controls."""
        # Plot type selection
        self.plot_type = widgets.Dropdown(
            options=[
                "Parameter vs Implausibility",
                "Pairwise Parameters",
                "Implausibility Distribution",
                "Parameter Correlation Heatmap",
                "3D Parameter Visualization",
                "Implausibility Radar",
                # "Bayesian Style Comparison",
            ],
            value="Parameter vs Implausibility",
            description="Plot Type:",
            style={"description_width": "initial"},
        )

        # Add observer to show/hide plot-specific controls
        self.plot_type.observe(self._update_visible_controls, names="value")

        # Parameter selection widgets
        self.param_x = widgets.Dropdown(
            options=self.param_names,
            value=self.param_names[0] if self.param_names else None,
            description="X Parameter:",
            disabled=False,
        )

        self.param_y = widgets.Dropdown(
            options=self.param_names,
            value=(
                self.param_names[1]
                if len(self.param_names) > 1
                else self.param_names[0]
            ),
            description="Y Parameter:",
            disabled=False,
        )

        self.param_z = widgets.Dropdown(
            options=self.param_names,
            value=(
                self.param_names[2]
                if len(self.param_names) > 2
                else self.param_names[0]
            ),
            description="Z Parameter:",
            disabled=False,
        )

        # Threshold slider
        self.threshold_slider = widgets.FloatSlider(
            value=self.threshold,
            min=0.5,
            max=10.0,
            step=0.1,
            description="Threshold:",
            continuous_update=False,
        )

        # Sample selection for radar plot
        self.sample_selector = widgets.IntSlider(
            value=0,
            min=0,
            max=len(self.samples_df) - 1,
            step=1,
            description="Sample Index:",
            continuous_update=False,
        )

        # Create parameter checkboxes
        self.param_checkboxes = []
        for param in self.param_names:
            cb = widgets.Checkbox(
                value=True,  # Default all selected
                description=param,
                disabled=False,
                indent=False,
            )
            self.param_checkboxes.append(cb)

        # Group checkboxes in a container with scroll
        self.param_checkbox_container = widgets.VBox(
            self.param_checkboxes,
            layout=widgets.Layout(
                width="auto", height="200px", overflow_y="auto", border="1px solid #ddd"
            ),
        )

        # Label for the checkbox group
        self.param_selection_label = widgets.Label("Select Parameters to Display:")

        # NROY filter
        self.nroy_filter = widgets.Checkbox(
            value=False, description="Show only NROY points", disabled=False
        )

        # Update button
        self.update_button = widgets.Button(
            description="Update Plot",
            button_style="primary",
            tooltip="Click to update the plot",
        )
        self.update_button.on_click(self._update_plot)

        # Output area for the plot
        self.output = widgets.Output()

        # Group controls for selective display
        self.param_selectors = widgets.HBox([self.param_x, self.param_y, self.param_z])
        # Container for the parameter selection controls
        self.param_selection_controls = widgets.VBox(
            [self.param_selection_label, self.param_checkbox_container]
        )

        self.radar_controls = widgets.HBox([self.sample_selector])

        # Main controls that are always visible
        controls_top = widgets.HBox([self.plot_type, self.threshold_slider])
        controls_bottom = widgets.HBox([self.nroy_filter])

        self.main_layout = widgets.VBox(
            [
                controls_top,
                self.param_selectors,
                self.radar_controls,
                controls_bottom,
                self.update_button,
                self.output,
            ]
        )

        # Initially hide plot-specific controls
        self.radar_controls.layout.display = "none"
        self.param_z.layout.display = "none"  # Initially hide Z parameter (only for 3D)
        self.param_selection_controls.layout.display = "none"  # Initially hidden

    def _update_visible_controls(self, change: dict):
        """Show/hide controls based on selected plot type."""
        plot_type = change["new"]

        # Default - show X and Y parameters, hide Z parameter
        self.param_x.layout.display = "inline-flex"
        self.param_y.layout.display = "inline-flex"
        self.param_z.layout.display = "none"
        self.param_selection_controls.layout.display = "none"

        # Hide all conditional controls by default
        self.radar_controls.layout.display = "none"

        # Default - hide NROY filter (hide for all plots initially)
        self.nroy_filter.layout.display = "none"

        # Show controls based on plot type
        if plot_type == "3D Parameter Visualization":
            # Show all three parameters for 3D
            self.param_z.layout.display = "inline-flex"
            # Show NROY filter for 3D viz
            self.nroy_filter.layout.display = "flex"

        elif plot_type == "Implausibility Radar":
            # Show sample selector for radar plot
            self.radar_controls.layout.display = "flex"

        elif plot_type in [
            "Parameter Correlation Heatmap",
            "Implausibility Distribution",
            # "Bayesian Style Comparison",
        ]:
            # Hide parameter selectors for plots that don't use them
            if plot_type in [
                "Parameter Correlation Heatmap",
                "Implausibility Distribution",
            ]:
                self.param_x.layout.display = "none"
                self.param_y.layout.display = "none"
            # Don't show NROY filter for these plots

        elif plot_type in [
            "Parameter vs Implausibility",
            "Pairwise Parameters",
            "Emulator Diagnostics",
        ]:
            # Show NROY filter only for these specific plots
            self.nroy_filter.layout.display = "flex"

    def _update_plot(self, _):
        """Update the plot based on current widget values."""
        with self.output:
            clear_output(wait=True)

            # Get current plot type
            plot_type = self.plot_type.value

            filtered_df = self.samples_df.copy()
            filtered_scores = self.impl_scores.copy()

            # Apply NROY filter if selected
            if self.nroy_filter.value:
                # NROY values are boolean
                nroy_mask = filtered_df["NROY"]
                filtered_df = filtered_df[nroy_mask].copy()

                # Filter implausibility scores accordingly
                if len(filtered_scores) == len(filtered_df.index.to_list()):
                    filtered_scores = filtered_scores[
                        np.array(nroy_mask.values, dtype=bool)
                    ].copy()

            # Update threshold
            threshold = self.threshold_slider.value
            filtered_df["NROY"] = (
                np.max(filtered_scores, axis=1) <= threshold
                if len(filtered_scores.shape) > 1
                else filtered_scores <= threshold
            )

            # Check if we have data to plot
            if len(filtered_df) == 0:
                plt.figure(figsize=(10, 6))
                plt.text(
                    0.5,
                    0.5,
                    "No data to display with current filters",
                    ha="center",
                    va="center",
                    fontsize=14,
                )
                plt.axis("off")
                plt.tight_layout()
                plt.show()
                return

            # Generate the selected plot
            assert isinstance(filtered_df, pd.DataFrame)
            assert isinstance(filtered_scores, NumpyLike)
            try:
                if plot_type == "Parameter vs Implausibility":
                    self._plot_parameter_vs_implausibility(filtered_df, filtered_scores)
                elif plot_type == "Pairwise Parameters":
                    self._plot_pairwise_parameters(filtered_df, filtered_scores)
                elif plot_type == "Implausibility Distribution":
                    self._plot_implausibility_distribution(filtered_scores)
                elif plot_type == "Parameter Correlation Heatmap":
                    self._plot_parameter_correlation(filtered_df)
                elif plot_type == "3D Parameter Visualization":
                    self._plot_3d_visualization(filtered_df, filtered_scores)
                elif plot_type == "Implausibility Radar":
                    self._plot_implausibility_radar(filtered_df, filtered_scores)
                # elif plot_type == "Bayesian Style Comparison":
                #     self._plot_bayesian_style_comparison(filtered_df, filtered_scores)
                plt.show()

            except Exception as e:
                plt.figure(figsize=(10, 6))
                plt.text(
                    0.5,
                    0.5,
                    f"Error generating plot: {e!s}",
                    ha="center",
                    va="center",
                    fontsize=14,
                )
                plt.axis("off")
                plt.tight_layout()
                plt.show()

    def _plot_parameter_vs_implausibility(
        self, df: pd.DataFrame, impl_scores: NumpyLike
    ):
        """Plot parameter vs implausibility."""
        threshold = self.threshold_slider.value
        param = self.param_x.value

        plt.figure(figsize=(12, 6))

        # Calculate minimum implausibility if we have multiple outputs
        if len(impl_scores.shape) > 1:
            min_impl = np.min(impl_scores, axis=1)
            max_impl = np.max(impl_scores, axis=1)
        else:
            min_impl = impl_scores
            max_impl = impl_scores

        # Create scatter plot
        plt.scatter(
            df[param],
            min_impl,
            c=max_impl,
            cmap="viridis_r",
            alpha=0.7,
            s=50,
            label="Min Implausibility",
        )

        # Add horizontal line for threshold
        plt.axhline(
            y=threshold, color="r", linestyle="--", label=f"Threshold = {threshold}"
        )

        plt.title(f"Parameter {param} vs Implausibility")
        plt.xlabel(param)
        plt.ylabel("Implausibility")
        plt.colorbar(label="Max Implausibility")
        plt.grid(True, alpha=0.3)
        plt.legend()
        plt.tight_layout()

    def _plot_pairwise_parameters(self, df: pd.DataFrame, impl_scores: NumpyLike):
        """Plot pairwise parameter visualization."""
        param_x = self.param_x.value
        param_y = self.param_y.value

        plt.figure(figsize=(10, 8))

        # Calculate maximum implausibility
        if len(impl_scores.shape) > 1:
            max_impl = np.max(impl_scores, axis=1)
        else:
            max_impl = impl_scores

        # Create scatter plot
        sc = plt.scatter(
            df[param_x], df[param_y], c=max_impl, cmap="viridis_r", alpha=0.7, s=50
        )

        # Highlight NROY points with an outline
        nroy_points = df[df["NROY"]]
        if not nroy_points.empty:
            plt.scatter(
                nroy_points[param_x],
                nroy_points[param_y],
                s=80,
                facecolors="none",
                edgecolors="g",
                linewidths=2,
                label="NROY Points",
            )

            # Add NROY region boundaries
            x_min, x_max = nroy_points[param_x].min(), nroy_points[param_x].max()
            y_min, y_max = nroy_points[param_y].min(), nroy_points[param_y].max()
            assert isinstance(x_min, float)
            assert isinstance(x_max, float)
            assert isinstance(y_min, float)
            assert isinstance(y_max, float)
            plt.axvline(x=x_min, color="g", linestyle="--", alpha=0.5)
            plt.axvline(x=x_max, color="g", linestyle="--", alpha=0.5)
            plt.axhline(y=y_min, color="g", linestyle="--", alpha=0.5)
            plt.axhline(y=y_max, color="g", linestyle="--", alpha=0.5)

        plt.title(f"Parameters {param_x} vs {param_y}")
        plt.xlabel(param_x)
        plt.ylabel(param_y)
        plt.colorbar(sc, label="Max Implausibility")
        plt.grid(True, alpha=0.3)
        if not nroy_points.empty:
            plt.legend()
        plt.tight_layout()

    def _plot_implausibility_distribution(self, impl_scores: NumpyLike):
        """Plot implausibility distribution."""
        threshold = self.threshold_slider.value

        plt.figure(figsize=(12, 6))

        # Flatten array if it's 2D (multiple outputs)
        if len(impl_scores.shape) > 1:
            # Plot distribution of maximum implausibility
            max_impl = np.max(impl_scores, axis=1)
            min_impl = np.min(impl_scores, axis=1)

            # Create histograms
            plt.hist(max_impl, bins=30, alpha=0.7, label="Max Implausibility")
            plt.hist(min_impl, bins=30, alpha=0.5, label="Min Implausibility")

            below_threshold = (max_impl <= threshold).sum() / len(max_impl) * 100
        else:
            # Single output case
            plt.hist(impl_scores, bins=30, alpha=0.7, label="Implausibility")
            below_threshold = (impl_scores <= threshold).sum() / len(impl_scores) * 100

        # Add vertical line for threshold
        plt.axvline(
            x=threshold, color="r", linestyle="--", label=f"Threshold = {threshold}"
        )

        plt.title(
            f"Implausibility Distribution\n"
            f"{below_threshold:.1f}% of points below threshold"
        )
        plt.xlabel("Implausibility")
        plt.ylabel("Frequency")
        plt.legend()
        plt.grid(True, alpha=0.3)
        plt.tight_layout()

    def _plot_parameter_correlation(self, df):
        """Plot parameter correlation heatmap."""
        # Get only parameter columns
        param_names = [p for p in self.param_names if p in df.columns]
        params_df = df[param_names]

        # Calculate correlation matrix
        corr = params_df.corr()

        # Create heatmap
        plt.figure(figsize=(12, 10))
        plt.imshow(corr, cmap="coolwarm", vmin=-1, vmax=1)

        # Add correlation values
        for i in range(len(corr)):
            for j in range(len(corr)):
                plt.text(
                    j,
                    i,
                    f"{corr.iloc[i, j]:.2f}",
                    ha="center",
                    va="center",
                    color="white" if abs(corr.iloc[i, j]) > 0.5 else "black",
                )

        # Add labels
        plt.xticks(np.arange(len(param_names)), param_names, rotation=45, ha="right")
        plt.yticks(np.arange(len(param_names)), param_names)

        plt.title("Parameter Correlation Heatmap")
        plt.colorbar(label="Correlation")
        plt.tight_layout()

    def _plot_3d_visualization(self, df: pd.DataFrame, impl_scores: NumpyLike):
        """Create a 3D visualization of parameters."""
        threshold = self.threshold_slider.value
        param_x = self.param_x.value
        param_y = self.param_y.value
        param_z = self.param_z.value

        # Calculate max implausibility
        if len(impl_scores.shape) > 1:
            max_impl = np.max(impl_scores, axis=1)
        else:
            max_impl = impl_scores

        # Create 3D plot
        fig = plt.figure(figsize=(12, 10))
        ax = fig.add_subplot(111, projection="3d")

        # Create scatter plot
        scatter = ax.scatter(
            df[param_x],
            df[param_y],
            df[param_z],
            c=max_impl,
            cmap="viridis_r",
            s=30,  # pyright: ignore[reportCallIssue]
            alpha=0.7,
        )

        # Highlight NROY points
        nroy_points = df[max_impl <= threshold]
        if not nroy_points.empty:
            ax.scatter(
                nroy_points[param_x],
                nroy_points[param_y],
                nroy_points[param_z],
                color="green",
                s=50,  # pyright: ignore[reportCallIssue]
                alpha=1.0,
                marker="o",
                label="NROY Points",
            )

        # Add labels
        ax.set_xlabel(param_x)
        ax.set_ylabel(param_y)
        ax.set_zlabel(param_z)  # pyright: ignore[reportAttributeAccessIssue]

        # Add colorbar
        cbar = fig.colorbar(scatter, ax=ax, pad=0.1)
        cbar.set_label("Implausibility")

        # Add threshold line on colorbar
        cbar.ax.axhline(y=threshold, color="r", linestyle="--")

        plt.title("3D Visualization of Parameters and Implausibility")

        if not nroy_points.empty:
            plt.legend()

        plt.tight_layout()

    def _plot_implausibility_radar(self, df: pd.DataFrame, impl_scores: NumpyLike):
        """Create radar plots showing implausibility for different outputs."""
        # Make sure impl_scores is 2D
        if len(impl_scores.shape) == 1:
            impl_scores = impl_scores.reshape(-1, 1)

        threshold = self.threshold_slider.value
        sample_idx = self.sample_selector.value

        # Make sure sample_idx is within range
        n_samples = len(df)
        if sample_idx >= n_samples:
            sample_idx = n_samples - 1
            self.sample_selector.value = sample_idx

        # Get values for the selected sample
        if sample_idx < len(impl_scores):
            sample_scores = impl_scores[sample_idx]

            # Get sample parameters
            sample_params = df.iloc[sample_idx].copy()

            # Remove non-parameter entries
            for key in ["min_implausibility", "max_implausibility", "NROY"]:
                if key in sample_params:
                    sample_params.pop(key)

            # Set up angles for radar plot
            n_outputs = len(self.output_names)
            angles = np.linspace(0, 2 * np.pi, n_outputs, endpoint=False).tolist()
            angles += angles[:1]  # Close the loop

            # Create extended scores array (for closing the loop)
            extended_scores = [*sample_scores.tolist(), sample_scores[0]]

            # Create figure
            fig, ax = plt.subplots(figsize=(12, 8), subplot_kw={"polar": True})

            # Plot threshold
            plt.plot(
                angles, [threshold] * len(angles), "r--", linewidth=1, label="Threshold"
            )

            # Fill the area below threshold
            plt.fill(angles, [threshold] * len(angles), "r", alpha=0.1)

            # Plot implausibility
            plt.plot(angles, extended_scores, "b-", linewidth=2, label="Implausibility")
            plt.fill(angles, extended_scores, "b", alpha=0.1)

            # Set labels
            plt.xticks(angles[:-1], self.output_names, fontsize=10)

            # Set y limits
            max_score = max(*extended_scores, threshold * 1.5)
            plt.ylim(0, max_score)

            # Create title with parameter values
            param_str = "\n".join(
                [f"{key}={value:.3g}" for key, value in sample_params.items()]
            )

            plt.title(
                f"Implausibility Radar for Sample {sample_idx}\n{param_str}",
                fontsize=12,
            )
            plt.legend(loc="upper right")

            plt.tight_layout()

            # Also show a bar chart with the same data
            plt.figure(figsize=(12, 6))
            bars = plt.bar(self.output_names, sample_scores, alpha=0.7)

            # Color bars based on threshold
            for i, bar in enumerate(bars):
                if sample_scores[i] > threshold:
                    bar.set_color("r")
                else:
                    bar.set_color("g")

            plt.axhline(
                y=threshold, color="r", linestyle="--", label=f"Threshold = {threshold}"
            )
            plt.xlabel("Output")
            plt.ylabel("Implausibility")
            plt.title(f"Implausibility for Sample {sample_idx}")
            plt.xticks(rotation=45, ha="right")
            plt.grid(True, alpha=0.3)
            plt.legend()
            plt.tight_layout()
        else:
            plt.figure(figsize=(10, 6))
            plt.text(
                0.5,
                0.5,
                "Sample index out of range",
                ha="center",
                va="center",
                fontsize=14,
            )
            plt.axis("off")
            plt.tight_layout()

    # def _plot_bayesian_style_comparison(
    #    self, df: pd.DataFrame, impl_scores: NumpyLike
    # ):
    #     """
    #     Create a Bayesian-style visualization showing parameter constraints
    #     with prior and posterior using existing dashboard controls.

    #     This matches the style shown in the example image with:
    #     - Gray shaded prior distributions
    #     - Blue histogram posterior (NROY) distributions
    #     - Support for LaTeX formatted parameter labels
    #     """
    #     import numpy as np

    #     # Calculate max implausibility for each sample
    #     if len(impl_scores.shape) > 1:
    #         max_impl = np.max(impl_scores, axis=1)
    #     else:
    #         max_impl = impl_scores

    #     # Get threshold for NROY classification
    #     threshold = self.threshold_slider.value

    #     # Create NROY indicator (these are our "posterior" samples)
    #     nroy_mask = max_impl <= threshold

    #     # Get the selected parameters from existing UI controls
    #     selected_params = [self.param_x.value, self.param_y.value]

    #     # Remove duplicates while preserving order
    #     selected_params = list(dict.fromkeys(selected_params))

    #     # Create the figure
    #     n_params = len(selected_params)
    #     n_cols = min(2, n_params)
    #     n_rows = (n_params + n_cols - 1) // n_cols

    #     fig = plt.figure(figsize=(6 * n_cols, 4 * n_rows))

    #     # Set the title
    #     title = "History Matching Results for Parameters"
    #     subtitle = "(using NROY points as posterior)"
    #     full_title = f"{title}\n{subtitle}"

    #     # Set the overall title if we have multiple plots
    #     if n_params > 1:
    #         fig.suptitle(full_title, fontsize=16, y=0.98)

    #     # Function to create nice parameter labels with LaTeX
    #     def format_param_label(param):
    #         # Format the parameter name nicely for display
    #         if "log" in param.lower():
    #             base_name = param.replace("log_", "").replace("log", "")
    #             return rf"$\mu_{{{base_name}}}$"
    #         if "_" in param:
    #             parts = param.split("_")
    #             if len(parts) == 2:
    #                 return f"$log_{{10}}({parts[0]}_{{v}}/{parts[0]}_{{h}})$"
    #             return param
    #         return param

    #     # Plot each parameter
    #     for i, param in enumerate(selected_params):
    #         ax = fig.add_subplot(n_rows, n_cols, i + 1)

    #         # Get the prior range (all samples)
    #         param_min = df[param].min()
    #         param_max = df[param].max()

    #         # Add padding
    #         padding = 0.1 * (param_max - param_min)
    #         param_min -= padding
    #         param_max += padding

    #         # Get the posterior data (NROY points)
    #         posterior_data = df.loc[nroy_mask, param]

    #         # Create bins
    #         bins = np.linspace(param_min, param_max, 20).tolist()

    #         # Plot prior (flat uniform distribution)
    #         prior_height = 0.4  # Height for the prior bar
    #         ax.fill_between(
    #             [param_min, param_max],
    #             [0, 0],
    #             [prior_height, prior_height],
    #             color="lightgray",
    #             alpha=0.5,
    #             label="Prior",
    #         )

    #         # Plot posterior
    #         if len(posterior_data) > 0:
    #             ax.hist(
    #                 posterior_data,
    #                 bins=bins,
    #                 density=True,
    #                 alpha=0.7,
    #                 color="royalblue",
    #                 label="Posterior",
    #             )

    #         # Set labels and limits
    #         ax.set_xlabel(format_param_label(param))
    #         ax.set_ylabel("Frequency")
    #         assert isinstance(param_min, float)
    #         assert isinstance(param_max, float)
    #         ax.set_xlim(param_min, param_max)

    #         # Show legend on first plot only
    #         if i == 0:
    #             ax.legend()

    #         # Set title only for single plot
    #         if n_params == 1:
    #             ax.set_title(full_title)

    #     plt.tight_layout()
    #     if n_params > 1:
    #         plt.subplots_adjust(top=0.9)  # Make room for suptitle

    def display(self):
        """Display the dashboard."""
        heading = widgets.HTML(value="<h2>History Matching Dashboard</h2>")

        # Display the heading and instructions first
        display(heading)
        display(self.main_layout)
        # Initialize the first plot
        self._update_plot(None)



import gpytorch
import torch
from gpytorch import ExactMarginalLogLikelihood
from gpytorch.distributions import MultitaskMultivariateNormal, MultivariateNormal
from gpytorch.kernels import MultitaskKernel, ScaleKernel
from gpytorch.likelihoods import MultitaskGaussianLikelihood
from gpytorch.means import MultitaskMean
from torch import optim
from torch.optim.lr_scheduler import LRScheduler

from autoemulate.callbacks.early_stopping import (
    EarlyStopping,
    EarlyStoppingException,
)
from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import (
    DeviceLike,
    GaussianLike,
    GaussianProcessLike,
    TensorLike,
)
from autoemulate.data.utils import set_random_seed
from autoemulate.emulators.base import GaussianProcessEmulator
from autoemulate.emulators.gaussian_process import (
    CovarModuleFn,
    MeanModuleFn,
)
from autoemulate.transforms.standardize import StandardizeTransform
from autoemulate.transforms.utils import make_positive_definite

from .kernel import (
    matern_3_2_kernel,
    matern_5_2_kernel,
    matern_5_2_plus_rq,
    rbf,
    rbf_plus_constant,
    rbf_plus_linear,
    rbf_times_linear,
    rq_kernel,
)
from .mean import constant_mean, linear_mean, poly_mean, zero_mean


class GaussianProcess(GaussianProcessEmulator, gpytorch.models.ExactGP):
    """
    Gaussian Process Emulator.

    This class implements an exact Gaussian Process emulator using the GPyTorch library

    It supports:
    - multi-task Gaussian processes
    - custom mean and kernel specification
    """

    # TODO: refactor to work more like PyTorchBackend once any subclasses implemented
    optimizer_cls: type[optim.Optimizer] = optim.AdamW
    optimizer: optim.Optimizer
    lr: float = 2e-1
    scheduler_cls: type[LRScheduler] | None = None

    def __init__(  # noqa: PLR0913 allow too many arguments since all currently required
        self,
        x: TensorLike,
        y: TensorLike,
        standardize_x: bool = False,
        standardize_y: bool = True,
        likelihood_cls: type[MultitaskGaussianLikelihood] = MultitaskGaussianLikelihood,
        mean_module_fn: MeanModuleFn = constant_mean,
        covar_module_fn: CovarModuleFn = rbf_plus_constant,
        posterior_predictive: bool = False,
        epochs: int = 50,
        lr: float = 2e-1,
        early_stopping: EarlyStopping | None = None,
        device: DeviceLike | None = None,
        **kwargs,
    ):
        """
        Initialize the GaussianProcess emulator.

        Parameters
        ----------
        x: TensorLike
            Input features, expected to be a 2D tensor of shape (n_samples, n_features).
        y: TensorLike
            Target values, expected to be a 2D tensor of shape (n_samples, n_tasks).
        likelihood_cls: type[MultitaskGaussianLikelihood]
            Likelihood class to use for the model. Defaults to
            `MultitaskGaussianLikelihood`.
        mean_module_fn: MeanModuleFn
            Function to create the mean module. Defaults to `constant_mean`.
        covar_module_fn: CovarModuleFn
            Function to create the covariance module. Defaults to `rbf`.
        posterior_predictive: bool
            If True, the model will return the posterior predictive distribution that
            by default includes observation noise (both global and task-specific). If
            False, it will return the posterior distribution over the modelled function.
            Defaults to False.
        epochs: int
            Number of training epochs. Defaults to 50.
        lr: float
            Learning rate for the optimizer. Defaults to 2e-1.
        early_stopping: EarlyStopping | None
            An optional EarlyStopping callback. Defaults to None.
        device: DeviceLike | None
            Device to run the model on. If None, uses the default device (usually CPU or
            GPU). Defaults to None.
        """
        # Init device
        TorchDeviceMixin.__init__(self, device=device)

        x, y = self._convert_to_tensors(x, y)
        x, y = self._move_tensors_to_device(x, y)

        # Local variables for number of features and tasks
        n_features = x.shape[1]
        num_tasks = y.shape[1]
        num_tasks_torch = torch.Size([num_tasks])

        # Initialize the mean and covariance modules
        mean_module = mean_module_fn(n_features, num_tasks_torch)
        covar_module = covar_module_fn(n_features, num_tasks_torch)

        # If the combined kernel is not a ScaleKernel, wrap it in one
        covar_module = (
            covar_module
            if isinstance(covar_module, ScaleKernel)
            else ScaleKernel(covar_module, batch_shape=num_tasks_torch)
        )

        # Init likelihood
        likelihood = likelihood_cls(num_tasks=num_tasks)
        likelihood = likelihood.to(self.device)

        # Init must be called with preprocessed data
        gpytorch.models.ExactGP.__init__(
            self, train_inputs=x, train_targets=y, likelihood=likelihood
        )
        self.likelihood = likelihood
        self.x_transform = StandardizeTransform() if standardize_x else None
        self.y_transform = StandardizeTransform() if standardize_y else None
        self.mean_module = mean_module
        self.covar_module = covar_module
        self.epochs = epochs
        self.lr = lr
        self.optimizer = self.optimizer_cls(self.parameters(), lr=self.lr)  # type: ignore[call-arg] since all optimizers include lr
        self.scheduler_setup(kwargs)
        self.early_stopping = early_stopping
        self.posterior_predictive = posterior_predictive
        self.num_tasks = num_tasks
        self.to(self.device)

    @staticmethod
    def is_multioutput():
        """GaussianProcess supports multioutput."""
        return True

    def forward(self, x: TensorLike):
        """Forward pass of the Gaussian Process model."""
        mean = self.mean_module(x)
        assert isinstance(mean, torch.Tensor)
        covar = self.covar_module(x)
        return MultitaskMultivariateNormal.from_batch_mvn(
            MultivariateNormal(mean, covar)
        )

    def _fit(self, x: TensorLike, y: TensorLike):
        self.train()
        self.likelihood.train()

        mll = ExactMarginalLogLikelihood(self.likelihood, self)

        # Set the training data in case changed since init
        self.set_train_data(x, y, strict=False)

        # Initialize early stopping
        if self.early_stopping is not None:
            self.early_stopping.on_train_begin()

        # Avoid `"epoch" is possibly unbound` type error at the end
        epoch = 0
        for epoch in range(self.epochs):
            self.optimizer.zero_grad()
            output = self(x)
            loss = mll(output, y)
            assert isinstance(loss, torch.Tensor)
            loss = -loss
            loss.backward()
            self.optimizer.step()

            if self.early_stopping is not None:
                try:
                    # TODO: use validation loss instead, see #589
                    self.early_stopping.on_epoch_end(self, epoch, loss.item())
                except EarlyStoppingException:
                    # EarlyStopping prints a message if this happens
                    break

            # Update learning rate if scheduler is defined
            if self.scheduler is not None:
                self.scheduler.step()

        if self.early_stopping is not None:
            self.early_stopping.on_train_end(self, epoch)

        # Update learning rate if scheduler is defined
        if self.scheduler is not None:
            self.scheduler.step()

    def _predict(self, x: TensorLike, with_grad: bool):
        with torch.set_grad_enabled(with_grad), gpytorch.settings.fast_pred_var():
            self.eval()
            self.likelihood.eval()
            x = x.to(self.device)

            num_points = x.shape[0]
            num_tasks = self.num_tasks

            # Use a heuristic for max batch size based on approximate memory usage
            # e.g. (batch_size * num_tasks) ** 2 * 4 (f32) ~ 100MB
            # and ensure batch size is at least 1
            max_batch_size = max(5000 // num_tasks, 1)

            # Lists to store batches of means and covs
            means_list = []
            covs_list = []

            # Loop over batches of points for predictionss
            for i in range(0, num_points, max_batch_size):
                x_batch = x[i : i + max_batch_size]

                # Get predictive output with full covariance between points and tasks
                output = self(x_batch)
                if self.posterior_predictive:
                    output = self.likelihood(output)
                assert isinstance(output, GaussianProcessLike)

                mean = output.mean
                cov = output.covariance_matrix
                assert isinstance(mean, TensorLike)
                assert isinstance(cov, TensorLike)

                num_batch = x_batch.shape[0]

                # Reshape mean
                mean = mean.reshape(num_batch, num_tasks)  # (num_batch, num_tasks)

                # Task covariance blocks for each point
                cov_blocks = cov.reshape(num_batch, num_tasks, num_batch, num_tasks)

                # Take diagonal along batch to only keep task covariance
                task_covs = cov_blocks[
                    torch.arange(num_batch), :, torch.arange(num_batch), :
                ]  # (num_batch, num_tasks, num_tasks)

                means_list.append(mean)
                covs_list.append(task_covs)

            # Concatenate batches
            means = torch.cat(means_list, dim=0)  # (num_points, num_tasks)
            assert means.shape == (num_points, num_tasks)
            covs = torch.cat(covs_list, dim=0)  # (num_points, num_tasks, num_tasks)
            assert covs.shape == (num_points, num_tasks, num_tasks)

            # Construct output distribution and ensure positive definiteness (with
            # jitter and clamping of eigvals) since removing inter-point covariance
            # from the output can affect postive definiteness
            output_distribution = GaussianLike(
                means,
                make_positive_definite(
                    covs, min_jitter=1e-6, max_tries=3, clamp_eigvals=True
                ),
            )
            assert output_distribution.batch_shape == torch.Size([num_points])
            assert output_distribution.event_shape == torch.Size([num_tasks])
            return output_distribution

    @staticmethod
    def get_tune_params():
        """Return the hyperparameters to tune for the Gaussian Process model."""
        scheduler_params = GaussianProcess.scheduler_params()
        return {
            "mean_module_fn": [
                constant_mean,
                zero_mean,
                linear_mean,
                poly_mean,
            ],
            "covar_module_fn": [
                rbf,
                matern_5_2_kernel,
                matern_3_2_kernel,
                rq_kernel,
                rbf_plus_constant,
                rbf_plus_linear,
                matern_5_2_plus_rq,
                rbf_times_linear,
            ],
            "epochs": [50, 100, 200],
            "lr": [5e-1, 1e-1, 5e-2, 1e-2],
            "likelihood_cls": [MultitaskGaussianLikelihood],
            "scheduler_cls": scheduler_params["scheduler_cls"],
            "scheduler_kwargs": scheduler_params["scheduler_kwargs"],
        }


class GaussianProcessCorrelated(GaussianProcess):
    """
    Multioutput exact GP implementation with correlated task covariance.

    This class extends the `GaussianProcess` to support correlated task covariance
    by using a `MultitaskKernel` with a rank-1 covariance factor and a `MultitaskMean`
    for the mean function.

    It is designed to handle multi-task Gaussian processes where the tasks are
    correlated, allowing for more flexible modeling of multi-output data.
    """

    # TODO: refactor the init as similar to exact GP base class
    def __init__(  # noqa: PLR0913 allow too many arguments since all currently required
        self,
        x: TensorLike,
        y: TensorLike,
        standardize_x: bool = False,
        standardize_y: bool = True,
        likelihood_cls: type[MultitaskGaussianLikelihood] = MultitaskGaussianLikelihood,
        mean_module_fn: MeanModuleFn = constant_mean,
        covar_module_fn: CovarModuleFn = rbf_plus_constant,
        posterior_predictive: bool = False,
        epochs: int = 50,
        lr: float = 2e-1,
        early_stopping: EarlyStopping | None = None,
        seed: int | None = None,
        device: DeviceLike | None = None,
        **kwargs,
    ):
        """
        Initialize the GaussianProcessCorrelated emulator.

        Parameters
        ----------
        x: TensorLike
            Input features, expected to be a 2D tensor of shape (n_samples, n_features).
        y: TensorLike
            Target values, expected to be a 2D tensor of shape (n_samples, n_tasks).
        likelihood_cls: type[MultitaskGaussianLikelihood]
            Likelihood class to use for the model. Defaults to
            `MultitaskGaussianLikelihood`.
        mean_module_fn: MeanModuleFn
            Function to create the mean module. Defaults to `constant_mean`.
        covar_module_fn: CovarModuleFn
            Function to create the covariance module. Defaults to `rbf`.
        posterior_predictive: bool
            If True, the model will return the posterior predictive distribution that
            by default includes observation noise (both global and task-specific). If
            False, it will return the posterior distribution over the modelled function.
            Defaults to False.
        epochs: int
            Number of training epochs.
        activation: type[nn.Module]
            Activation function to use in the model. Defaults to `nn.ReLU`.
        lr: float
            Learning rate for the optimizer. Defaults to 2e-1.
        early_stopping: EarlyStopping | None
            An optional EarlyStopping callback. Defaults to None.
        seed: int | None
            Random seed for reproducibility. If None, no seed is set. Defaults to None.
        device: DeviceLike | None
            Device to run the model on. If None, uses the default device (usually CPU or
            GPU). Defaults to None.
        """
        # Init device
        TorchDeviceMixin.__init__(self, device=device)

        if seed is not None:
            set_random_seed(seed)

        # Convert to 2D tensors if needed and move to device
        x, y = self._move_tensors_to_device(*self._convert_to_tensors(x, y))

        # Initialize the mean and covariance modules
        n_features = tuple(x.shape)[1]
        mean_module = mean_module_fn(n_features, None)
        covar_module = covar_module_fn(n_features, None)

        # Mean and covariance modules for multitask
        num_tasks = tuple(y.shape)[1]
        mean_module = MultitaskMean(mean_module, num_tasks=num_tasks)
        covar_module = MultitaskKernel(covar_module, num_tasks=num_tasks, rank=1)

        # TODO: identify if initialization of task covariance factor is needed for
        # deterministic behavior
        # with torch.no_grad():
        #     mean_module.task_covar_factor.fill_(1.0)  # type: ignore  # noqa: PGH003
        #     covar_module.task_covar_factor.fill_(1.0)  # type: ignore  # noqa: PGH003

        # Init likelihood
        likelihood = likelihood_cls(num_tasks=num_tasks)
        likelihood = likelihood.to(self.device)

        # Init must be called with preprocessed data
        gpytorch.models.ExactGP.__init__(
            self,
            train_inputs=x,
            train_targets=y,
            likelihood=likelihood,
        )
        self.x_transform = StandardizeTransform() if standardize_x else None
        self.y_transform = StandardizeTransform() if standardize_y else None
        self.likelihood = likelihood
        self.mean_module = mean_module
        self.covar_module = covar_module
        self.epochs = epochs
        self.lr = lr
        self.optimizer = self.optimizer_cls(self.parameters(), lr=self.lr)  # type: ignore[call-arg] since all optimizers include lr
        self.scheduler_setup(kwargs)
        self.early_stopping = early_stopping
        self.posterior_predictive = posterior_predictive
        self.num_tasks = num_tasks
        self.to(self.device)

    def forward(self, x):
        """Forward pass of the Gaussian Process Correlated model."""
        mean_x = self.mean_module(x)
        assert isinstance(mean_x, TensorLike)
        covar_x = self.covar_module(x)
        return GaussianProcessLike(mean_x, covar_x)
